# SPDX-License-Identifier: LGPL-3.0-or-later
import os

import torch

torch.set_num_threads(1)
torch.set_num_interop_threads(1)
# testing purposes; device should always be set explicitly
torch.set_default_device("cuda:9999999")

# Make sure DDP uses correct device if applicable
LOCAL_RANK = os.environ.get("LOCAL_RANK")
LOCAL_RANK = int(0 if LOCAL_RANK is None else LOCAL_RANK)

if os.environ.get("DEVICE") == "cpu" or torch.cuda.is_available() is False:
    DEVICE = torch.device("cpu")
else:
    DEVICE = torch.device(f"cuda:{LOCAL_RANK}")

SEED = 1024



# SPDX-License-Identifier: LGPL-3.0-or-later
"""
Base force module for torch-admp.

This module provides the abstract base class for force modules in torch-admp,
which defines the common interface for force calculations. It includes
standardized input validation and unit conversion functionality.
"""

from abc import ABC, abstractmethod
from typing import Dict, Optional, Tuple

import torch

from torch_admp.utils import TorchConstants


class BaseForceModule(torch.nn.Module, ABC):
    """
    Abstract base class for force modules in torch-admp.

    This class provides a common interface for force modules that take atomic
    positions and simulation box as input and return energy values. It is designed
    to be compatible with OpenMM-torch and sets up a constants library as a
    class attribute for necessary physical constants.

    Notes
    -----
    All subclasses must implement the _forward_impl method to define specific force
    calculations.
    """

    def __init__(self, units_dict: Optional[Dict] = None) -> None:
        """
        Initialize the BaseForceModule.

        Parameters
        ----------
        units_dict : Optional[Dict], default=None
            Dictionary containing unit conversion factors. If None, default units
            will be used.

        Attributes
        ----------
        const_lib : TorchConstants
            Library containing physical constants and unit conversions.
        """
        torch.nn.Module.__init__(self)
        self.const_lib = TorchConstants(units_dict)

    def forward(
        self,
        positions: torch.Tensor,
        box: Optional[torch.Tensor],
        pairs: torch.Tensor,
        ds: torch.Tensor,
        buffer_scales: torch.Tensor,
        params: Dict[str, torch.Tensor],
    ) -> torch.Tensor:
        """
        Compute the potential energy for the given atomic configuration.

        This method validates input dimensions and then calls the abstract
        _forward_impl method which must be implemented by subclasses.

        Parameters
        ----------
        positions : torch.Tensor
            Atomic positions with shape (natoms, 3). Each row contains the
            x, y, z coordinates of an atom.
        box : Optional[torch.Tensor]
            Simulation box vectors with shape (3, 3). Each row represents a
            box vector. Required for periodic boundary conditions.
        pairs : torch.Tensor
            Tensor of atom pairs with shape (n_pairs, 2). Each row contains
            the indices of two atoms that form a pair.
        ds : torch.Tensor
            Distance tensor with shape (n_pairs,). Contains the distances
            between atom pairs specified in the pairs tensor.
        buffer_scales : torch.Tensor
            Buffer scales for each pair with shape (n_pairs,). Contains values
            of 1 if i < j else 0 for each pair, used for buffer management.
        params : Dict[str, torch.Tensor]
            Dictionary of parameters for the PES model. Common parameters include
            atomic charges, Lennard-Jones parameters, etc.

        Returns
        -------
        torch.Tensor
            Scalar energy tensor representing the total potential energy of the
            system.
        """
        # verify and standardize shape of input tensors
        (
            _positions,
            _box,
            _pairs,
            _ds,
            _buffer_scales,
        ) = self.standardize_input_tensor(
            positions,
            box,
            pairs,
            ds,
            buffer_scales,
        )
        nf = _positions.size(0)
        _params = {k: v.reshape(nf, -1) for k, v in params.items()}

        # Call the implementation in subclasses
        # Use getattr to explicitly access the tensor and avoid type checker issues
        length_coeff = getattr(self.const_lib, "length_coeff")
        return self._forward_impl(
            _positions * length_coeff,
            _box * length_coeff if _box is not None else None,
            _pairs,
            _ds * length_coeff,
            _buffer_scales,
            _params,
        )

    @abstractmethod
    @torch.jit.export
    def _forward_impl(
        self,
        positions: torch.Tensor,
        box: Optional[torch.Tensor],
        pairs: torch.Tensor,
        ds: torch.Tensor,
        buffer_scales: torch.Tensor,
        params: Dict[str, torch.Tensor],
    ) -> torch.Tensor:
        """
        Implementation of the potential energy calculation.

        This abstract method must be implemented by subclasses to define the
        specific potential energy surface (PES) model.

        Parameters
        ----------
        positions : torch.Tensor
            Atomic positions with shape (nframes, natoms, 3). Each frame contains
            the x, y, z coordinates of all atoms.
        box : Optional[torch.Tensor]
            Simulation box vectors with shape (nframes, 3, 3) or None if input was None.
            Each frame contains three box vectors. Required for periodic boundary conditions.
        pairs : torch.Tensor
            Tensor of atom pairs with shape (nframes, n_pairs, 2). Each frame contains
            the indices of two atoms that form a pair.
        ds : torch.Tensor
            Distance tensor with shape (nframes, n_pairs). Contains the distances
            between atom pairs specified in the pairs tensor for each frame.
        buffer_scales : torch.Tensor
            Buffer scales for each pair with shape (nframes, n_pairs). Contains values
            of 1 if i < j else 0 for each pair, used for buffer management.
        params : Dict[str, torch.Tensor]
            Dictionary of parameters for the PES model. Common parameters include
            atomic charges, Lennard-Jones parameters, etc.

        Returns
        -------
        torch.Tensor
            Scalar energy tensor representing the total potential energy of the
            system.

        Raises
        ------
        NotImplementedError
            If called directly on the base class. Must be implemented by
            subclasses.
        """

    def standardize_input_tensor(
        self,
        positions: torch.Tensor,
        box: Optional[torch.Tensor],
        pairs: torch.Tensor,
        ds: torch.Tensor,
        buffer_scales: torch.Tensor,
    ) -> Tuple[
        torch.Tensor, Optional[torch.Tensor], torch.Tensor, torch.Tensor, torch.Tensor
    ]:
        """
        Verify the shape of input tensors.

        Parameters
        ----------
        positions : torch.Tensor
            Atomic positions with shape (natoms, 3) or (nframes, natoms, 3)
        box : Optional[torch.Tensor]
            Simulation box vectors with shape (3, 3) or (nframes, 3, 3)
        pairs : torch.Tensor
            Tensor of atom pairs with shape (n_pairs, 2) or (nframes, n_pairs, 2)
        ds : torch.Tensor
            Distance tensor with shape (n_pairs,) or (nframes, n_pairs)
        buffer_scales : torch.Tensor
            Buffer scales with shape (n_pairs,) or (nframes, n_pairs)

        Returns
        -------
        Tuple[torch.Tensor, Optional[torch.Tensor], torch.Tensor, torch.Tensor, torch.Tensor]
            Standardized tensors with shapes:
            - positions: (nframes, natoms, 3)
            - box: (nframes, 3, 3) or None if input was None
            - pairs: (nframes, n_pairs, 2)
            - ds: (nframes, n_pairs)
            - buffer_scales: (nframes, n_pairs)

        Raises
        ------
        ValueError
            If any tensor has incorrect dimensions
        """
        nframes = 1
        # Check positions dimensions
        if positions.dim() == 3:
            # Batched input: (nframes, natoms, 3)
            if positions.size(2) != 3:
                raise ValueError(
                    f"positions must have shape (nframes, natoms, 3), got {positions.shape}"
                )
            nframes = positions.size(0)
        elif positions.dim() == 2:
            # Single system: (natoms, 3)
            if positions.size(1) != 3:
                raise ValueError(
                    f"positions must have shape (natoms, 3), got {positions.shape}"
                )
            positions = positions.unsqueeze(0)
        else:
            raise ValueError(
                f"positions must be 2D or 3D tensor, got {positions.dim()}D"
            )

        # Check box dimensions if provided
        if box is not None:
            if box.dim() == 3:
                # Batched input: (nframes, 3, 3)
                if box.shape[1:] != (3, 3):
                    raise ValueError(
                        f"box must have shape (nframes, 3, 3), got {box.shape}"
                    )
                if box.size(0) != nframes:
                    raise ValueError(
                        f"box is expected to have {nframes} frame(s), got {box.size(0)}"
                    )
            elif box.dim() == 2:
                # Single system: (3, 3)
                if box.shape != (3, 3):
                    raise ValueError(f"box must have shape (3, 3), got {box.shape}")
                if nframes != 1:
                    raise ValueError(
                        f"box must include a frame dimension when positions has {nframes} frames"
                    )
                box = box.unsqueeze(0)
            else:
                raise ValueError(f"box must be 2D or 3D tensor, got {box.dim()}D")

        # Check pairs dimensions
        if pairs.dim() == 3:
            # Batched input: (nframes, n_pairs, 2)
            if pairs.size(2) != 2:
                raise ValueError(
                    f"pairs must have shape (nframes, n_pairs, 2), got {pairs.shape}"
                )
            if pairs.size(0) != nframes:
                raise ValueError(
                    f"pairs is expected to have {nframes} frame(s), got {pairs.size(0)}"
                )
        elif pairs.dim() == 2:
            # Single system: (n_pairs, 2)
            if pairs.size(1) != 2:
                raise ValueError(
                    f"pairs must have shape (n_pairs, 2), got {pairs.shape}"
                )
            if nframes != 1:
                raise ValueError(
                    f"pairs must include a frame dimension when positions has {nframes} frames"
                )
            pairs = pairs.unsqueeze(0)
        else:
            raise ValueError(f"pairs must be 2D or 3D tensor, got {pairs.dim()}D")
        n_pairs = pairs.size(-2)

        # Check ds dimensions
        if ds.dim() == 2:
            # Batched input: (nframes, n_pairs)
            if ds.size(0) != nframes:
                raise ValueError(
                    f"ds is expected to have {nframes} frame(s), got {ds.size(0)}"
                )
            if ds.size(1) != n_pairs:
                raise ValueError(
                    f"ds is expected to have {n_pairs} pairs(s), got {ds.size(1)}"
                )
        elif ds.dim() == 1:
            # Single system: (n_pairs,)
            if ds.size(0) != n_pairs:
                raise ValueError(
                    f"ds is expected to have {n_pairs} pairs(s), got {ds.size(0)}"
                )
            if nframes != 1:
                raise ValueError(
                    f"ds must include a frame dimension when positions has {nframes} frames"
                )
            ds = ds.unsqueeze(0)
        else:
            raise ValueError(f"ds must be 1D or 2D tensor, got {ds.dim()}D")

        # Check buffer_scales dimensions
        if buffer_scales.dim() == 2:
            # Batched input: (nframes, n_pairs)
            if buffer_scales.size(0) != nframes:
                raise ValueError(
                    f"buffer_scales is expected to have {nframes} frame(s), got {buffer_scales.size(0)}"
                )
            if buffer_scales.size(1) != n_pairs:
                raise ValueError(
                    f"buffer_scales is expected to have {n_pairs} pairs(s), got {buffer_scales.size(1)}"
                )
        elif buffer_scales.dim() == 1:
            # Single system: (n_pairs,)
            if buffer_scales.size(0) != n_pairs:
                raise ValueError(
                    f"buffer_scales is expected to have {n_pairs} pairs(s), got {buffer_scales.size(0)}"
                )
            if nframes != 1:
                raise ValueError(
                    f"buffer_scales must include a frame dimension when positions has {nframes} frames"
                )
            buffer_scales = buffer_scales.unsqueeze(0)
        else:
            raise ValueError(
                f"buffer_scales must be 1D or 2D tensor, got {buffer_scales.dim()}D"
            )

        return positions, box, pairs, ds, buffer_scales



# SPDX-License-Identifier: LGPL-3.0-or-later
import numpy as np
from ase import Atoms, io

n_atoms = 100
l_box = 10.0

charges = np.random.uniform(-1.0, 1.0, n_atoms)
charges -= np.mean(charges)
positions = np.random.uniform(0.0, l_box, (n_atoms, 3))

atoms = Atoms(
    n_atoms * "H", positions=positions, charges=charges, cell=[l_box, l_box, l_box]
)
io.write("coord.xyz", atoms)
io.write("system.data", atoms, format="lammps-data", atom_style="full")



# SPDX-License-Identifier: LGPL-3.0-or-later
"""
Particle Mesh Ewald (PME) implementation for torch-admp.

This module implements the Coulomb energy calculation using the Particle Mesh Ewald
method, which splits the calculation into real-space and reciprocal-space
contributions for improved efficiency in periodic systems. It includes support for
slab corrections and various optimization methods.
"""

import math
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from scipy import special

from torch_admp.base_force import BaseForceModule
from torch_admp.env import DEVICE, GLOBAL_PT_FLOAT_PRECISION
from torch_admp.recip import bspline, setup_kpts, setup_kpts_integer, spread_charges
from torch_admp.utils import safe_inverse


class CoulombForceModule(BaseForceModule):
    """
    Coulomb energy module with Particle Mesh Ewald (PME).

    This module implements the Coulomb energy calculation using the Particle Mesh Ewald
    method, which splits the calculation into real-space and reciprocal-space
    contributions for improved efficiency in periodic systems.

    Parameters
    ----------
    rcut : float
        Real-space cutoff distance
    ethresh : float, optional
        Energy threshold for PME accuracy, by default 1e-5
    kspace : bool, optional
        Whether to include reciprocal space contribution, by default True
    rspace : bool, optional
        Whether to include real space contribution, by default True
    slab_corr : bool, optional
        Whether to apply slab correction, by default False
    slab_axis : int, optional
        Axis at which the slab correction is applied, by default 2
    units_dict : Optional[Dict], optional
        Dictionary of unit conversions, by default None
    sel : Optional[list[int]], optional
        Selection list for neighbor list, by default None
    kappa : Optional[float], optional
        Inverse screening length [Å^-1], by default None
    spacing : Optional[List[float]], optional
        Grid spacing for reciprocal space, by default None
    """

    def __init__(
        self,
        rcut: float,
        ethresh: float = 1e-5,
        kspace: bool = True,
        rspace: bool = True,
        slab_corr: bool = False,
        slab_axis: int = 2,
        units_dict: Optional[Dict] = None,
        sel: Optional[list[int]] = None,
        kappa: Optional[float] = None,
        spacing: Union[List[float], float, None] = None,
    ) -> None:
        """
        Initialize the CoulombForceModule with PME.

        Parameters
        ----------
        rcut : float
            Real-space cutoff distance
        ethresh : float, optional
            Energy threshold for PME accuracy, by default 1e-5
        kspace : bool, optional
            Whether to include reciprocal space contribution, by default True
        rspace : bool, optional
            Whether to include real space contribution, by default True
        slab_corr : bool, optional
            Whether to apply slab correction, by default False
        slab_axis : int, optional
            Axis at which the slab correction is applied, by default 2
        units_dict : Optional[Dict], optional
            Dictionary of unit conversions, by default None
        sel : Optional[list[int]], optional
            Selection list for neighbor list, by default None
        kappa : Optional[float], optional
            Inverse screening length [Å^-1], by default None
        spacing : Optional[List[float]], optional
            Grid spacing for reciprocal space, by default None
        """
        BaseForceModule.__init__(self, units_dict)

        self.kspace_flag = kspace
        if kappa is not None:
            self.kappa = kappa
        else:
            if self.kspace_flag:
                kappa = math.sqrt(-math.log(2 * ethresh)) / rcut
                self.kappa = kappa / getattr(self.const_lib, "length_coeff")
            else:
                self.kappa = 0.0
        self.ethresh = ethresh
        self.kmesh = torch.ones(3, dtype=torch.long, device=DEVICE)
        if spacing is not None:
            if isinstance(spacing, float):
                spacing = [spacing, spacing, spacing]
            self.spacing = torch.tensor(
                np.array(spacing), dtype=GLOBAL_PT_FLOAT_PRECISION, device=DEVICE
            )
        else:
            self.spacing = spacing
        self.rspace_flag = rspace
        self.slab_corr_flag = slab_corr
        self.slab_axis = slab_axis

        self.real_energy = torch.tensor(
            0.0, device=DEVICE, dtype=GLOBAL_PT_FLOAT_PRECISION
        )
        self.reciprocal_energy = torch.tensor(
            0.0, device=DEVICE, dtype=GLOBAL_PT_FLOAT_PRECISION
        )
        self.self_energy = torch.tensor(
            0.0, device=DEVICE, dtype=GLOBAL_PT_FLOAT_PRECISION
        )
        self.non_neutral_energy = torch.tensor(
            0.0, device=DEVICE, dtype=GLOBAL_PT_FLOAT_PRECISION
        )
        self.slab_corr_energy = torch.tensor(
            0.0, device=DEVICE, dtype=GLOBAL_PT_FLOAT_PRECISION
        )

        # Currently only supprots pme_order=6
        # Because only the 6-th order spline function is hard implemented
        self.pme_order: int = 6
        n_mesh = int(self.pme_order**3)

        # global variables for the reciprocal module, all related to pme_order
        bspline_range = torch.arange(
            -self.pme_order // 2, self.pme_order // 2, device=DEVICE
        )
        shift_y, shift_x, shift_z = torch.meshgrid(
            bspline_range, bspline_range, bspline_range, indexing="ij"
        )
        self.pme_shifts = (
            torch.stack((shift_x, shift_y, shift_z))
            .transpose(0, 3)
            .reshape((1, n_mesh, 3))
        )

        self.rcut = rcut
        self.sel = sel

    def get_rcut(self) -> float:
        """
        Get the cutoff radius.

        Returns
        -------
        float
            Cutoff radius
        """
        return self.rcut

    def get_sel(self) -> Optional[list[int]]:
        """
        Get `sel` list of DP model.

        Returns
        -------
        Optional[list[int]]
            The number of selected neighbors for each type of atom.
        """
        return self.sel

    def _forward_impl(
        self,
        positions: torch.Tensor,
        box: Optional[torch.Tensor],
        pairs: torch.Tensor,
        ds: torch.Tensor,
        buffer_scales: torch.Tensor,
        params: Dict[str, torch.Tensor],
    ) -> torch.Tensor:
        """Coulomb energy model with PME algorithm for KSpace

        Parameters
        ----------
        positions : torch.Tensor
            Atomic positions with shape (nframes, natoms, 3). Each frame contains
            x, y, z coordinates of all atoms.
        box : Optional[torch.Tensor]
            Simulation box vectors with shape (nframes, 3, 3) or None if input was None.
            Each frame contains three box vectors. Required for periodic boundary conditions.
        pairs : torch.Tensor
            Tensor of atom pairs with shape (nframes, n_pairs, 2). Each frame contains
            the indices of two atoms that form a pair.
        ds : torch.Tensor
            Distance tensor with shape (nframes, n_pairs). Contains the distances
            between atom pairs specified in the pairs tensor for each frame.
        buffer_scales : torch.Tensor
            Buffer scales for each pair with shape (nframes, n_pairs). Contains values
            of 1 if i < j else 0 for each pair, used for buffer management.
        params : Dict[str, torch.Tensor]
            Dictionary of parameters for the Coulomb model:
            {"charge": t_charges} # atomic charges with shape (nframes, natoms).

        Returns
        -------
        energy: torch.Tensor
            Scalar energy tensor for single system or (nframes,) for batched systems
            representing the total Coulomb energy.
        """
        # nframes, natoms,
        nf = positions.size(0)
        na = positions.size(1)
        charges = params["charge"].reshape(nf, na)

        if box is not None:
            energy = self._forward_pbc(
                charges, positions, box, pairs, ds, buffer_scales
            )
        else:
            energy = self._forward_obc(charges, pairs, ds, buffer_scales)
        return energy / getattr(self.const_lib, "energy_coeff")

    def _forward_pbc(
        self,
        charges: torch.Tensor,
        positions: torch.Tensor,
        box: torch.Tensor,
        pairs: torch.Tensor,
        ds: torch.Tensor,
        buffer_scales: torch.Tensor,
    ) -> torch.Tensor:
        """
        Calculate Coulomb energy for periodic boundary conditions.

        Parameters
        ----------
        charges : torch.Tensor
            Atomic charges
        positions : torch.Tensor
            Atomic positions
        box : torch.Tensor
            Simulation box vectors
        pairs : torch.Tensor
            Tensor of atom pairs
        ds : torch.Tensor
            Distance tensor
        buffer_scales : torch.Tensor
            Buffer scales for each pair

        Returns
        -------
        torch.Tensor
            Total Coulomb energy
        """
        if self.rspace_flag:
            self.real_energy = self._forward_pbc_real(charges, pairs, ds, buffer_scales)
        if self.kspace_flag:
            self.reciprocal_energy = self._forward_pbc_reciprocal(
                charges, positions, box
            )
            self.self_energy = self._forward_pbc_self(charges)
            self.non_neutral_energy = self._forward_pbc_non_neutral(charges, box)
        if self.slab_corr_flag:
            self.slab_corr_energy = self._forward_slab_corr(charges, positions, box, ds)
        coul_energy = (
            self.real_energy
            + self.reciprocal_energy
            + self.self_energy
            + self.non_neutral_energy
            + self.slab_corr_energy
        )
        return coul_energy

    def _forward_pbc_real(
        self,
        charges: torch.Tensor,
        pairs: torch.Tensor,
        ds: torch.Tensor,
        buffer_scales: torch.Tensor,
    ) -> torch.Tensor:
        """
        Calculate real-space contribution to Coulomb energy.

        Parameters
        ----------
        charges : torch.Tensor
            Atomic charges
        pairs : torch.Tensor
            Tensor of atom pairs
        ds : torch.Tensor
            Distance tensor
        buffer_scales : torch.Tensor
            Buffer scales for each pair

        Returns
        -------
        torch.Tensor
            Real-space contribution to Coulomb energy
        """
        # qi or qj: nf, np
        qi = torch.gather(charges, 1, pairs[:, :, 0])
        qj = torch.gather(charges, 1, pairs[:, :, 1])

        e_sr = torch.sum(
            torch.erfc(self.kappa * ds)
            * qi
            * qj
            * safe_inverse(ds, threshold=1e-4)
            * buffer_scales,
            dim=-1,
        ) * getattr(self.const_lib, "dielectric")
        return e_sr

    def _forward_pbc_reciprocal(
        self,
        charges: torch.Tensor,
        positions: torch.Tensor,
        box: torch.Tensor,
    ) -> torch.Tensor:
        """
        Calculate reciprocal-space contribution to Coulomb energy.

        Parameters
        ----------
        charges : torch.Tensor
            Atomic charges
        positions : torch.Tensor
            Atomic positions
        box : torch.Tensor
            Simulation box vectors

        Returns
        -------
        torch.Tensor
            Reciprocal-space contribution to Coulomb energy
        """
        device = positions.device
        nf = positions.size(0)

        box_inv = torch.linalg.inv(box)
        volume = torch.det(box)
        box_diag = torch.diagonal(box, dim1=1, dim2=2)
        if self.spacing is not None:
            spacing = torch.as_tensor(
                self.spacing, dtype=box_diag.dtype, device=box_diag.device
            )
            self.kmesh = torch.ceil(box_diag / spacing).to(torch.long)
        else:
            self.kmesh = torch.ceil(
                2 * self.kappa * box_diag / (3.0 * self.ethresh ** (1.0 / 5.0))
            ).to(torch.long)

        # for electrostatic, exclude gamma point
        gamma_flag = False
        coeff_k_func = _coeff_k_1
        all_ener = []
        for ii in range(nf):
            # charges: -1, 1
            _charges = charges[ii].reshape(-1, 1)
            # mapping charges onto mesh
            meshed_charges = spread_charges(
                positions[ii],
                box_inv[ii],
                _charges,
                self.kmesh[ii],
                self.pme_shifts,
                self.pme_order,
            )
            kpts_int = setup_kpts_integer(self.kmesh[ii])
            kpts = setup_kpts(box_inv[ii], kpts_int)
            m = torch.linspace(
                -self.pme_order // 2 + 1,
                self.pme_order // 2 - 1,
                self.pme_order - 1,
                device=device,
            ).reshape(self.pme_order - 1, 1, 1)
            theta_k = torch.prod(
                torch.sum(
                    bspline(m + self.pme_order / 2)
                    * torch.cos(
                        2
                        * torch.pi
                        * m
                        * kpts_int[None]
                        / self.kmesh[ii].float().reshape(1, 1, 3)
                    ),
                    dim=0,
                ),
                dim=1,
            )

            S_k = torch.fft.fftn(meshed_charges).flatten()
            if not gamma_flag:
                coeff_k = coeff_k_func(kpts[3, 1:], self.kappa, volume[ii])
                E_k = coeff_k * (
                    (S_k[1:].real ** 2 + S_k[1:].imag ** 2) / theta_k[1:] ** 2
                )
            else:
                coeff_k = coeff_k_func(kpts[3, :], self.kappa, volume[ii])
                E_k = coeff_k * ((S_k.real**2 + S_k.imag**2) / theta_k**2)
            # return
            all_ener.append(torch.sum(E_k) * getattr(self.const_lib, "dielectric"))
        return torch.stack(all_ener)

    def _forward_pbc_self(
        self,
        charges: torch.Tensor,
    ) -> torch.Tensor:
        r"""
        -\frac{\alpha}{\sqrt{\pi}} \sum_{i} q_i^2
        """
        if self.kspace_flag:
            coeff = self.kappa / getattr(self.const_lib, "sqrt_pi")
            return -torch.sum(coeff * charges**2, dim=-1) * getattr(
                self.const_lib, "dielectric"
            )
        else:
            return torch.zeros(charges.size(0), device=charges.device)

    def _forward_pbc_non_neutral(
        self,
        charges: torch.Tensor,
        box: torch.Tensor,
    ) -> torch.Tensor:
        """
        Calculate non-neutral correction for Coulomb energy.

        Parameters
        ----------
        charges : torch.Tensor
            Atomic charges
        box : torch.Tensor
            Simulation box vectors

        Returns
        -------
        torch.Tensor
            Non-neutral correction to Coulomb energy
        """
        volume = torch.det(box)
        # total charge
        Q_tot = torch.sum(charges, dim=-1)

        coeff = (
            -getattr(self.const_lib, "pi")
            / (2 * volume * self.kappa**2)
            * getattr(self.const_lib, "dielectric")
        )
        e_corr_non = coeff * Q_tot**2
        return e_corr_non

    def _forward_slab_corr(
        self,
        charges: torch.Tensor,
        positions: torch.Tensor,
        box: torch.Tensor,
        ds: torch.Tensor,
    ) -> torch.Tensor:
        r"""
        Slab correction energy (ref: 10.1063/1.3216473)

        E = \frac{2\pi}{V} \varepsilon \left( M_z^2 - Q_{\text{tot}} \sum_i q_i z_i + \frac{Q_{\text{tot}}^2 L_z^2}{12} \right)
        """
        volume = torch.det(box)
        pre_corr = (
            2
            * getattr(self.const_lib, "pi")
            / volume
            * getattr(self.const_lib, "dielectric")
        )
        # dipole moment in axis direction
        Mz = torch.sum(charges * positions[:, :, self.slab_axis], dim=-1)
        # total charge
        Q_tot = torch.sum(charges, dim=-1)
        # length of the box in axis direction
        Lz = torch.norm(box[:, self.slab_axis], dim=-1)

        e_corr = pre_corr * (
            Mz**2
            - Q_tot
            * (torch.sum(charges * positions[:, :, self.slab_axis] ** 2, dim=-1))
            - torch.pow(Q_tot, 2) * torch.pow(Lz, 2) / 12
        )
        return e_corr

    def _forward_obc(
        self,
        charges: torch.Tensor,
        pairs: torch.Tensor,
        ds: torch.Tensor,
        buffer_scales: torch.Tensor,
    ) -> torch.Tensor:
        """
        Calculate Coulomb energy for open boundary conditions.

        Parameters
        ----------
        charges : torch.Tensor
            Atomic charges
        pairs : torch.Tensor
            Tensor of atom pairs
        ds : torch.Tensor
            Distance tensor
        buffer_scales : torch.Tensor
            Buffer scales for each pair

        Returns
        -------
        torch.Tensor
            Coulomb energy for open boundary conditions
        """
        # qi or qj: nf, np
        qi = torch.gather(charges, 1, pairs[:, :, 0])
        qj = torch.gather(charges, 1, pairs[:, :, 1])
        ds_inv = safe_inverse(ds)
        E_inter = qi * qj * getattr(self.const_lib, "dielectric") * ds_inv
        coul_energy = torch.sum(E_inter * buffer_scales, dim=-1)
        return coul_energy


def setup_ewald_parameters(
    rcut: float,
    box: Union[torch.Tensor, np.ndarray, None] = None,
    threshold: float = 1e-5,
    spacing: Optional[float] = None,
    method: str = "openmm",
) -> Tuple[float, int, int, int]:
    """
    Given the cutoff distance, and the required precision, determine the parameters used in
    Ewald sum, including: kappa, kx, ky, and kz.

    Parameters
    ----------
    rcut : float
        Cutoff distance
    threshold : float
        Expected average relative errors in force
    box : torch.Tensor or np.ndarray
        Lattice vectors in (3 x 3) matrix
        Keep unit consistent with rcut
    spacing : float, optional
        Fourier spacing to determine K, used in gromacs method
        Keep unit consistent with rcut
    method : str
        Method to determine ewald parameters.
        Valid values: "openmm" or "gromacs".
        If openmm, the algorithm can refer to http://docs.openmm.org/latest/userguide/theory/02_standard_forces.html#coulomb-interaction-with-particle-mesh-ewald
        If gromacs, the algorithm is adapted from gromacs source code

    Returns
    -------
    kappa: float
        Ewald parameter, in 1/lenght unit
    kx, ky, kz: int
        number of the k-points mesh
    """
    if box is None:
        return 0.1, 1, 1, 1

    if isinstance(box, torch.Tensor):
        box = torch.Tensor.numpy(box, force=True)

    # assert orthogonal box
    assert (
        np.inner(box[0], box[1]) == 0.0
    ), "Only orthogonal box is supported currently."
    assert (
        np.inner(box[0], box[2]) == 0.0
    ), "Only orthogonal box is supported currently."
    assert (
        np.inner(box[1], box[2]) == 0.0
    ), "Only orthogonal box is supported currently."

    if method == "openmm":
        kappa = np.sqrt(-np.log(2 * threshold)) / rcut
        kx = np.ceil(2 * kappa * box[0, 0] / (3.0 * threshold ** (1.0 / 5.0))).astype(
            int
        )
        ky = np.ceil(2 * kappa * box[1, 1] / (3.0 * threshold ** (1.0 / 5.0))).astype(
            int
        )
        kz = np.ceil(2 * kappa * box[2, 2] / (3.0 * threshold ** (1.0 / 5.0))).astype(
            int
        )
    elif method == "gromacs":
        assert spacing is not None, "Spacing must be provided for gromacs method."
        # determine kappa
        kappa = 5.0
        i = 0
        while special.erfc(kappa * rcut) > threshold:
            i += 1
            kappa *= 2

        n = i + 60
        low = 0.0
        high = kappa
        for k in range(n):
            kappa = (low + high) / 2
            if special.erfc(kappa * rcut) > threshold:
                low = kappa
            else:
                high = kappa
        # determine K
        kx = np.ceil(box[0, 0] / spacing).astype(int)
        ky = np.ceil(box[1, 1] / spacing).astype(int)
        kz = np.ceil(box[2, 2] / spacing).astype(int)
    else:
        raise ValueError(
            f"Invalid method: {method}." "Valid methods: 'openmm', 'gromacs'"
        )

    return kappa, kx, ky, kz


def _coeff_k_1(
    ksq: torch.Tensor,
    kappa: float,
    volume: torch.Tensor,
):
    """
    Calculate coefficient for k-space contribution.

    Parameters
    ----------
    ksq : torch.Tensor
        Square of k-vectors
    kappa : float
        Eald parameter
    volume : torch.Tensor
        Volume of the simulation box

    Returns
    -------
    torch.Tensor
        Coefficient for k-space contribution
    """
    return 2 * torch.pi / volume / ksq * torch.exp(-ksq / 4 / kappa**2)



# SPDX-License-Identifier: LGPL-3.0-or-later
"""
Optimization algorithms for torch-admp.

This module implements various optimization algorithms used for charge equilibration
and other optimization tasks in the torch-admp package, including line search,
conjugate gradient methods, and other optimization utilities.
"""

from typing import Callable

import torch


@torch.jit.unused
def line_search(
    func_value: Callable,
    func_grads: Callable,
    x0: torch.Tensor,
    eps: float = 1e-6,
    fk: torch.Tensor = None,
    gk: torch.Tensor = None,
    pk: torch.Tensor = None,
    **kwargs,
) -> torch.Tensor:
    """
    Perform line search to find optimal step size.

    Parameters
    ----------
    func_value : Callable
        Function to compute the value of the objective function
    func_grads : Callable
        Function to compute gradients of the objective function
    x0 : torch.Tensor
        Initial point
    eps : float, optional
        Convergence threshold, by default 1e-6
    fk : torch.Tensor, optional
        Function value at x0, by default None
    gk : torch.Tensor, optional
        Gradient at x0, by default None
    pk : torch.Tensor, optional
        Search direction, by default None
    **kwargs
        Additional keyword arguments passed to func_value and func_grads

    Returns
    -------
    torch.Tensor
        Optimal point found by line search
    """
    history_x = torch.arange(3, dtype=x0.dtype, device=x0.device)
    if fk is None:
        x0 = x0.detach()
        x0.requires_grad = True
        if x0.grad is not None:
            x0.grad.detach_()
            x0.grad.zero_()
        fk = func_value(x0, **kwargs)
        gk = func_grads(fk, x0)
        pk = -gk / torch.norm(gk)
    history_f = [fk]

    xk = x0.detach()
    # xk.requires_grad = True
    for _ in range(2):
        if torch.norm(gk) / xk.shape[0] < eps:
            return xk
        xk = xk + pk
        xk.detach_()
        xk.requires_grad = True
        if xk.grad is not None:
            xk.grad.detach_()
            xk.grad.zero_()
        fk = func_value(xk, **kwargs)
        gk = func_grads(fk, xk)
        fk.detach_()
        history_f.append(fk)

    coeff_matrix = torch.stack(
        [history_x**2, history_x, torch.ones_like(history_x)], dim=1
    )
    y = torch.stack(history_f)
    coeff = torch.linalg.solve(coeff_matrix, y)
    # print(coeff[0])
    x_opt = x0 - coeff[1] / (2 * coeff[0]) * pk
    return x_opt


@torch.jit.unused
def quadratic_optimize(
    func_value: Callable,
    func_grads: Callable,
    xk: torch.Tensor,
    eps: float = 1e-4,
    ls_eps: float = 1e-4,
    max_iter: int = 20,
    **kwargs,
):
    """
    Perform quadratic optimization with conjugate gradient method.

    Parameters
    ----------
    func_value : Callable
        Function to compute the value of the objective function
    func_grads : Callable
        Function to compute gradients of the objective function
    xk : torch.Tensor
        Initial point
    eps : float, optional
        Convergence threshold, by default 1e-4
    ls_eps : float, optional
        Line search threshold, by default 1e-4
    max_iter : int, optional
        Maximum number of iterations, by default 20
    **kwargs
        Additional keyword arguments passed to func_value and func_grads

    Returns
    -------
    tuple
        Tuple containing (xk, fk, gk, converge_iter) where:
        - xk: optimal point
        - fk: function value at optimal point
        - gk: gradient at optimal point
        - converge_iter: iteration at which convergence was achieved
    """
    converge_iter: int = -1

    if xk.grad is not None:
        xk.grad.detach_()
        xk.grad.zero_()

    fk = func_value(xk, **kwargs)
    gk = func_grads(fk, xk)
    pk = -gk / torch.norm(gk)
    for ii in range(max_iter):
        fk.detach_()
        pk.detach_()
        # Selecting the step length
        x_new = line_search(func_value, func_grads, xk, ls_eps, fk, gk, pk, **kwargs)
        x_new.detach_()
        x_new.requires_grad = True
        if x_new.grad is not None:
            x_new.grad.detach_()
            x_new.grad.zero_()
        fk_new = func_value(x_new, **kwargs)
        gk_new = func_grads(fk_new, x_new)

        xk = x_new
        fk = fk_new

        norm_grad = torch.norm(gk_new) / xk.shape[0]
        # print(norm_grad)
        if norm_grad < eps:
            gk = gk_new
            converge_iter = ii
            break
        else:
            pk = update_pr(gk, pk, gk_new)
            gk = gk_new

    return xk, fk, gk, converge_iter


# class QuadraticProjectedGradientOptimizer(Optimizer):
#     def __init__(
#         self,
#         params,
#         lr: float = 1e-2,
#         tol: float = 1e-4,
#         max_iter: int = 20,
#         constraint_matrix: torch.Tensor = None,
#     ):
#         defaults = dict(lr=lr, tol=tol, max_iter=max_iter)
#         Optimizer.__init__(self, params, defaults)
#         self.constraint_matrix = constraint_matrix

#     def value_func(self, x: torch.Tensor) -> torch.Tensor:
#         return torch.empty_like(x)

#     def grad_func(self, x: torch.Tensor) -> torch.Tensor:
#         pass

#     def line_search(
#         self,
#         _x0: torch.Tensor,
#         positions: torch.Tensor,
#         box: Optional[torch.Tensor] = None,
#         fk: torch.Tensor = None,
#         gk: torch.Tensor = None,
#         pk: torch.Tensor = None,
#     ) -> torch.Tensor:
#         x0 = _x0.detach()
#         x0.requires_grad = True

#         history_x = torch.arange(3, dtype=torch.float64, device=positions.device)
#         if fk is None:
#             fk = self.value_func(x0, positions, box)
#             gk = self.func_grads(fk, x0)
#             pk = -gk / torch.norm(gk)
#         history_f = [fk]
#         xk = x0.detach()
#         xk.requires_grad = True
#         for _ in range(2):
#             if torch.norm(gk) / x0.shape[0] < self.ls_eps:
#                 return xk
#             xk = xk + pk
#             fk = self.func_energy(xk, positions, box)
#             gk = self.func_grads(fk, xk)
#             history_f.append(fk)

#         coeff_matrix = torch.stack(
#             [history_x**2, history_x, torch.ones_like(history_x)], dim=1
#         )
#         y = torch.stack(history_f)
#         coeff = torch.linalg.solve(coeff_matrix, y)
#         print(coeff)
#         x_opt = x0 - coeff[1] / (2 * coeff[0]) * pk
#         return x_opt


# class ConjugateGradientQuadraticOptimizer(Optimizer):
#     def __init__(self, params, lr=1e-2, tol=1e-6, max_iter=20):
#         defaults = dict(lr=lr, tol=tol, max_iter=max_iter)
#         Optimizer.__init__(self, params, defaults)

#     # The step method for the optimizer, fully compatible with torch.jit.script
#     def step(self, grads: torch.Tensor, params: list):
#         for group in self.param_groups:
#             lr = group["lr"]
#             tol = group["tol"]
#             max_iter = group["max_iter"]

#             # Initialization for conjugate gradient
#             r = -grads
#             p = r.clone()
#             r_norm = r.dot(r)

#             # Iterative conjugate gradient solver
#             for i in range(max_iter):
#                 # Compute the product of A and p
#                 Ap = torch.autograd.grad(p.dot(grads), params, retain_graph=True)
#                 Ap_flat = torch.cat([a.view(-1) for a in Ap])

#                 # Compute alpha
#                 alpha = r_norm / p.dot(Ap_flat)

#                 # Update parameters using alpha and search direction p
#                 for param, direction in zip(
#                     params, p.split([param.numel() for param in params])
#                 ):
#                     param.data.add_(lr * alpha * direction.view_as(param))

#                 r_new = r + alpha * Ap_flat
#                 r_new_norm = r_new.dot(r_new)

#                 if r_new_norm < tol:
#                     break

#                 beta = r_new_norm / r_norm
#                 p = r_new + beta * p
#                 r = r_new
#                 r_norm = r_new_norm

#     def zero_grad(self):
#         for group in self.param_groups:
#             for p in group["params"]:
#                 if p.grad is not None:
#                     p.grad.detach_()
#                     p.grad.zero_()


# # Define a simple model with the optimizer
# class SimpleModelWithOptimizer(nn.Module):
#     def __init__(self):
#         super(SimpleModelWithOptimizer, self).__init__()
#         self.fc = torch.nn.Linear(10, 1)
#         self.optimizer = JITConjugateGradient(self.fc.parameters(), lr=1e-2)

#     def forward(self, x, target):
#         output = self.fc(x)
#         loss = torch.nn.functional.mse_loss(output, target)
#         return loss

#     def optimize(self, x, target):
#         loss = self.forward(x, target)
#         self.optimizer.zero_grad()
#         grads = torch.autograd.grad(loss, self.fc.parameters(), create_graph=True)
#         grads_flat = torch.cat([g.view(-1) for g in grads])
#         self.optimizer.step(grads_flat, list(self.fc.parameters()))
#         return loss


# # Script the model
# model = SimpleModelWithOptimizer()

# # Dummy input and target for demonstration
# x = torch.randn(100, 10)
# target = torch.randn(100, 1)

# # Optimize the model
# for i in range(100):
#     loss = model.optimize(x, target)
#     print(f"Iteration {i}, Loss: {loss.item()}")

# # Now we can script the entire model
# scripted_model = torch.jit.script(model)

# # Saving and loading the scripted model
# scripted_model.save("model_with_optimizer.pt")
# loaded_model = torch.jit.load("model_with_optimizer.pt")

# # You can now use the loaded model in the same way


def update_sd(
    gk: torch.Tensor,
    pk: torch.Tensor,
    gk_new: torch.Tensor,
) -> torch.Tensor:
    """
    Update search direction using Steepest Descent Algorithm.

    Parameters
    ----------
    gk : torch.Tensor
        Current gradient
    pk : torch.Tensor
        Current search direction
    gk_new : torch.Tensor
        New gradient

    Returns
    -------
    torch.Tensor
        Updated search direction
    """
    gk = gk_new
    # Selection of the direction of the steepest descent
    pk = -gk / torch.linalg.norm(gk)
    return pk


def update_fr(
    gk: torch.Tensor,
    pk: torch.Tensor,
    gk_new: torch.Tensor,
) -> torch.Tensor:
    """
    Update search direction using Fletcher-Reeves Algorithm.

    Parameters
    ----------
    gk : torch.Tensor
        Current gradient
    pk : torch.Tensor
        Current search direction
    gk_new : torch.Tensor
        New gradient

    Returns
    -------
    torch.Tensor
        Updated search direction
    """
    old_gk = gk
    gk = gk_new
    # Line (16) of the Fletcher-Reeves algorithm
    chi = torch.linalg.norm(gk) ** 2 / torch.linalg.norm(old_gk) ** 2
    # Updated descent direction
    pk = -gk + chi * pk
    return pk


def update_pr(
    gk: torch.Tensor,
    pk: torch.Tensor,
    gk_new: torch.Tensor,
) -> torch.Tensor:
    """
    Update search direction using Polak-Ribiere Algorithm.

    Parameters
    ----------
    gk : torch.Tensor
        Current gradient
    pk : torch.Tensor
        Current search direction
    gk_new : torch.Tensor
        New gradient

    Returns
    -------
    torch.Tensor
        Updated search direction
    """
    old_gk = gk
    gk = gk_new
    # Line (16) of the Polak-Ribiere Algorithm
    chi = (gk - old_gk).dot(gk) / torch.linalg.norm(old_gk) ** 2
    chi = torch.where(chi > 0, chi, torch.zeros_like(chi))
    # Updated descent direction
    pk = -gk + chi * pk
    return pk


def update_hs(
    gk: torch.Tensor,
    pk: torch.Tensor,
    gk_new: torch.Tensor,
) -> torch.Tensor:
    """
    Update search direction using Hestenes-Stiefel Algorithm.

    Parameters
    ----------
    gk : torch.Tensor
        Current gradient
    pk : torch.Tensor
        Current search direction
    gk_new : torch.Tensor
        New gradient

    Returns
    -------
    torch.Tensor
        Updated search direction
    """
    old_gk = gk
    gk = gk_new
    chi = gk.dot(gk - old_gk) / pk.dot(gk - old_gk)
    # Updated descent direction
    pk = -gk + chi * pk
    return pk


def update_dy(
    gk: torch.Tensor,
    pk: torch.Tensor,
    gk_new: torch.Tensor,
) -> torch.Tensor:
    """
    Update search direction using Dai-Yuan Algorithm.

    Parameters
    ----------
    gk : torch.Tensor
        Current gradient
    pk : torch.Tensor
        Current search direction
    gk_new : torch.Tensor
        New gradient

    Returns
    -------
    torch.Tensor
        Updated search direction
    """
    old_gk = gk
    gk = gk_new
    chi = torch.linalg.norm(gk) ** 2 / pk.dot(gk - old_gk)
    # Updated descent direction
    pk = -gk + chi * pk
    return pk


def update_hz(
    gk: torch.Tensor,
    pk: torch.Tensor,
    gk_new: torch.Tensor,
) -> torch.Tensor:
    """
    Update search direction using Hager-Zhang Algorithm.

    Parameters
    ----------
    gk : torch.Tensor
        Current gradient
    pk : torch.Tensor
        Current search direction
    gk_new : torch.Tensor
        New gradient

    Returns
    -------
    torch.Tensor
        Updated search direction
    """
    old_gk = gk
    gk = gk_new
    delta_gk = gk - old_gk
    m = delta_gk - 2 * pk * torch.linalg.norm(delta_gk) ** 2 / pk.dot(delta_gk)
    n = gk / pk.dot(delta_gk)
    chi = m.dot(n)
    # Updated descent direction
    pk = -gk + chi * pk
    return pk



# SPDX-License-Identifier: LGPL-3.0-or-later
import unittest

import numpy as np
import torch
from scipy import constants

try:
    import jax
    import jax.numpy as jnp
    from dmff.admp.qeq import E_site3, E_sr3

    DMFF_AVAILABLE = True
except ImportError:
    DMFF_AVAILABLE = False

from torch_admp.nblist import TorchNeighborList
from torch_admp.qeq import (
    GaussianDampingForceModule,
    QEqForceModule,
    SiteForceModule,
    pgrad_optimize,
)
from torch_admp.utils import (
    calc_grads,
    calc_pgrads,
    to_numpy_array,
    vector_projection_coeff_matrix,
)

from . import SEED

# Generator with a fixed seed for reproducibility
rng = np.random.default_rng(seed=SEED)


class JaxTestData:
    def __init__(self):
        self.rcut = 5.0
        self.l_box = 20.0
        self.n_atoms = 100

        charges = rng.uniform(-1.0, 1.0, (self.n_atoms))
        self.charges = charges - charges.mean()
        self.positions = rng.random((self.n_atoms, 3)) * self.l_box
        self.box = np.diag([self.l_box, self.l_box, self.l_box])
        self.chi = np.ones(self.n_atoms)
        self.hardness = np.zeros(self.n_atoms)
        self.eta = np.ones(self.n_atoms) * 0.5

        # kJ/mol to eV
        j2ev = constants.physical_constants["joule-electron volt relationship"][0]
        # kJ/mol to eV/particle
        self.energy_coeff = j2ev * constants.kilo / constants.Avogadro


@unittest.skipIf(not DMFF_AVAILABLE, "dmff package not installed")
class TestGaussianDampingForceModule(unittest.TestCase, JaxTestData):
    def setUp(self) -> None:
        JaxTestData.__init__(self)

        self.module = GaussianDampingForceModule()
        self.jit_module = torch.jit.script(self.module)

    def test_consistent(self):
        positions = torch.tensor(self.positions, requires_grad=True)
        box = torch.tensor(self.box)
        charges = torch.tensor(self.charges)
        eta = torch.tensor(self.eta)

        nblist = TorchNeighborList(cutoff=self.rcut)
        pairs = nblist(positions, box)
        ds = nblist.get_ds()
        buffer_scales = nblist.get_buffer_scales()

        ener_pt = self.module(
            positions,
            box,
            pairs,
            ds,
            buffer_scales,
            {"charge": charges, "eta": eta},
        )
        force_pt = -calc_grads(ener_pt, positions)

        # in DMFF they use eta as sqrt(2) * Gaussian width
        jax_out = jax.value_and_grad(E_sr3, argnums=0)(
            jnp.array(self.positions),
            jnp.array(self.box),
            jnp.array(to_numpy_array(pairs)),
            jnp.array(self.charges),
            jnp.array(self.eta * np.sqrt(2.0)),
            jnp.array(to_numpy_array(buffer_scales)),
            True,
        )
        ener_jax = jax_out[0] * self.energy_coeff
        force_jax = -jax_out[1] * self.energy_coeff

        # energy [eV]
        np.testing.assert_allclose(
            to_numpy_array(ener_pt),
            ener_jax,
            atol=1e-6,
            rtol=1e-6,
        )
        # force [eV/A]
        np.testing.assert_allclose(
            to_numpy_array(force_pt).reshape(-1, 3),
            force_jax.reshape(-1, 3),
            atol=1e-6,
            rtol=1e-6,
        )


@unittest.skipIf(not DMFF_AVAILABLE, "dmff package not installed")
class TestSiteForceModule(unittest.TestCase, JaxTestData):
    def setUp(self) -> None:
        JaxTestData.__init__(self)

        self.module = SiteForceModule()
        self.jit_module = torch.jit.script(self.module)

    def test_consistent(self):
        positions = torch.tensor(self.positions, requires_grad=True)
        box = torch.tensor(self.box)
        charges = torch.tensor(self.charges)
        chi = torch.tensor(self.chi)
        hardness = torch.tensor(self.hardness)

        nblist = TorchNeighborList(cutoff=self.rcut)
        pairs = nblist(positions, box)
        ds = nblist.get_ds()
        buffer_scales = nblist.get_buffer_scales()

        ener_pt = self.module(
            positions,
            box,
            pairs,
            ds,
            buffer_scales,
            {"charge": charges, "chi": chi, "hardness": hardness},
        )

        ener_jax = E_site3(
            jnp.array(self.chi),
            jnp.array(self.hardness),
            jnp.array(self.charges),
        )

        # energy [eV]
        np.testing.assert_allclose(
            to_numpy_array(ener_pt),
            ener_jax,
            atol=1e-6,
            rtol=1e-6,
        )


@unittest.skipIf(not DMFF_AVAILABLE, "dmff package not installed")
class TestQEqForceModule(unittest.TestCase):
    """
    self consistent test (matrix inversion vs pgrad)
    """

    def setUp(self) -> None:
        self.rcut = 5.0
        self.l_box = 20.0
        self.ethresh = 1e-6
        self.n_atoms = 100

        positions = rng.random((self.n_atoms, 3)) * self.l_box
        self.positions = torch.tensor(positions, requires_grad=True)
        self.box = torch.tensor(np.diag([self.l_box, self.l_box, self.l_box]))
        charges = rng.uniform(-1.0, 1.0, (self.n_atoms))
        charges -= charges.mean()
        self.charges = torch.tensor(charges, requires_grad=True)

        chi = rng.random((self.n_atoms,))
        self.chi = torch.tensor(chi)
        self.hardness = torch.zeros(self.n_atoms)
        self.eta = torch.ones(self.n_atoms) * 0.5

        self.constraint_matrix = torch.ones((1, self.n_atoms))
        self.constraint_vals = torch.zeros(1)
        self.coeff_matrix = vector_projection_coeff_matrix(self.constraint_matrix)

        self.module_matinv = QEqForceModule(self.rcut, self.ethresh)
        self.module_pgrad = QEqForceModule(
            self.rcut, self.ethresh, eps=1e-6, ls_eps=1e-6, max_iter=100
        )
        self.jit_module = torch.jit.script(self.module_pgrad)

    def test_consistent(self):
        n_atoms = self.positions.shape[0]

        nblist = TorchNeighborList(cutoff=self.rcut)
        pairs = nblist(self.positions, self.box)
        ds = nblist.get_ds()
        buffer_scales = nblist.get_buffer_scales()

        # energy, q_opt
        out_matinv = self.module_matinv.solve_matrix_inversion(
            self.positions,
            self.box,
            self.chi,
            self.hardness,
            self.eta,
            pairs,
            ds,
            buffer_scales,
            self.constraint_matrix,
            self.constraint_vals,
        )
        forces_matinv = -calc_grads(out_matinv[0], self.positions)

        for method in ["lbfgs", "quadratic"]:
            out_pgrad = self.module_pgrad.solve_pgrad(
                self.charges,
                self.positions,
                self.box,
                self.chi,
                self.hardness,
                self.eta,
                pairs,
                ds,
                buffer_scales,
                self.constraint_matrix,
                self.constraint_vals,
                self.coeff_matrix,
                reinit_q=True,
                method=method,
            )
            forces_pgrad = -calc_grads(out_pgrad[0], self.positions)

            # convergence check
            assert self.module_pgrad.converge_iter >= 0

            pgrad = calc_pgrads(
                out_pgrad[0], out_pgrad[1], self.constraint_matrix, self.coeff_matrix
            )
            assert (pgrad.norm() / n_atoms).item() < self.module_pgrad.eps

            # energy [eV]
            np.testing.assert_allclose(
                to_numpy_array(out_matinv[0]),
                to_numpy_array(out_pgrad[0]),
                atol=1e-6,
                rtol=1e-6,
            )
            # force [eV/A]
            np.testing.assert_allclose(
                to_numpy_array(forces_matinv),
                to_numpy_array(forces_pgrad),
                atol=5e-6,
                rtol=5e-6,
            )
            # charge [e]
            np.testing.assert_allclose(
                to_numpy_array(out_matinv[1]),
                to_numpy_array(out_pgrad[1]),
                atol=5e-6,
                rtol=5e-6,
            )

        for method in ["lbfgs", "quadratic"]:
            out_jit = pgrad_optimize(
                self.jit_module,
                self.charges,
                self.positions,
                self.box,
                self.chi,
                self.hardness,
                self.eta,
                pairs,
                ds,
                buffer_scales,
                self.constraint_matrix,
                self.constraint_vals,
                self.coeff_matrix,
                reinit_q=True,
                method=method,
            )
            forces_jit = -calc_grads(out_jit[0], self.positions)

            # convergence check
            assert self.jit_module.converge_iter >= 0

            pgrad = calc_pgrads(
                out_jit[0], out_jit[1], self.constraint_matrix, self.coeff_matrix
            )
            assert (pgrad.norm() / n_atoms).item() < self.jit_module.eps

            # energy [eV]
            np.testing.assert_allclose(
                to_numpy_array(out_matinv[0]),
                to_numpy_array(out_jit[0]),
                atol=1e-6,
                rtol=1e-6,
            )
            # force [eV/A]
            np.testing.assert_allclose(
                to_numpy_array(forces_matinv),
                to_numpy_array(forces_jit),
                atol=5e-6,
                rtol=5e-6,
            )
            # charge [e]
            np.testing.assert_allclose(
                to_numpy_array(out_matinv[1]),
                to_numpy_array(out_jit[1]),
                atol=5e-6,
                rtol=5e-6,
            )

    def test_hessian(self):
        charges = rng.uniform(-1.0, 1.0, (self.n_atoms))
        charges -= charges.mean()
        charges = torch.tensor(charges)

        params = {
            "charge": charges,
            "eta": self.eta,
        }

        nblist = TorchNeighborList(cutoff=self.rcut)
        pairs = nblist(self.positions, self.box)
        ds = nblist.get_ds()
        buffer_scales = nblist.get_buffer_scales()

        hessian = self.module_matinv.calc_hessian(
            self.positions,
            self.box,
            self.chi,
            torch.zeros_like(self.chi),
            self.eta,
            pairs,
            ds,
            buffer_scales,
        )

        e1 = self.module_matinv.submodels["coulomb"](
            self.positions, self.box, pairs, ds, buffer_scales, params
        )
        e2 = self.module_matinv.submodels["damping"](
            self.positions, self.box, pairs, ds, buffer_scales, params
        )

        hessian = to_numpy_array(hessian)
        charges = to_numpy_array(charges)
        np.testing.assert_allclose(
            0.5 * np.inner(np.matmul(charges, hessian), charges),
            to_numpy_array(e1 + e2),
        )



# SPDX-License-Identifier: LGPL-3.0-or-later
"""Tests for neighbor list functionality in torch-admp.

This module contains tests to verify the correctness of neighbor list
implementations, including comparisons with reference implementations
and consistency checks between different neighbor list methods.
"""

import unittest

import freud
import numpy as np
import torch

from torch_admp.nblist import (
    TorchNeighborList,
    dp_nblist,
    extend_input_and_build_neighbor_list,
    sort_pairs,
    vesin_nblist,
)
from torch_admp.utils import to_numpy_array

from . import SEED


class TestTorchNeighborList(unittest.TestCase):
    """Test class for TorchNeighborList functionality.

    Tests that the TorchNeighborList correctly identifies atom pairs
    within the cutoff distance, comparing against freud reference data.
    """

    def setUp(self):
        """Set up test data for TorchNeighborList testing.

        Creates reference neighbor list data using freud and initializes
        TorchNeighborList with test positions and box.
        """
        # Set random generators with SEED for reproducibility
        np_rng = np.random.default_rng(SEED)
        # torch_rng = torch.Generator().manual_seed(SEED)

        # reference data
        rcut = 4.0
        l_box = 10.0
        box = np.diag([l_box, l_box, l_box])
        positions = np_rng.random((20, 3)) * l_box

        fbox = freud.box.Box.from_matrix(box)
        aq = freud.locality.AABBQuery(fbox, positions)
        res = aq.query(positions, dict(r_max=rcut, exclude_ii=True))
        nblist = res.toNeighborList()
        nblist = np.vstack((nblist[:, 0], nblist[:, 1])).T
        nblist = nblist.astype(np.int32)
        msk = (nblist[:, 0] - nblist[:, 1]) < 0
        self.nblist_ref = nblist[msk]

        self.nblist = TorchNeighborList(rcut)
        self.positions = torch.tensor(positions)
        self.box = torch.tensor(box)

    def test_pairs(self):
        """Check that pairs are in the neighbor list.

        Verifies that all pairs identified by TorchNeighborList are
        present in the reference neighbor list from freud.
        """
        pairs = self.nblist(self.positions, self.box)
        pairs = to_numpy_array(pairs)
        mask = pairs[:, 0] < pairs[:, 1]
        assert len(pairs[mask]) == len(self.nblist_ref)
        for p in pairs[mask]:
            mask = (self.nblist_ref[:, 0] == p[0]) & (self.nblist_ref[:, 1] == p[1])
            self.assertTrue(mask.any())


class TestNBList(unittest.TestCase):
    """Test class for neighbor list consistency.

    Tests consistency between different neighbor list implementations
    to ensure they produce identical results.
    """

    def setUp(self) -> None:
        """Set up test data for neighbor list consistency testing.

        Creates random atomic positions in a periodic box and initializes
        parameters for neighbor list testing.
        """
        # Set random generators with SEED for reproducibility
        np_rng = np.random.default_rng(SEED)
        # torch_rng = torch.Generator().manual_seed(SEED)

        # reference data
        l_box = 10.0
        box = np.diag([l_box, l_box, l_box])
        positions = np_rng.random((100, 3)) * l_box

        self.positions = torch.tensor(positions)
        self.box = torch.tensor(box)

        # test: cutoff > l_box / 2!!!
        self.rcut = 6.0
        self.nnei = 150

    @unittest.skipIf(
        extend_input_and_build_neighbor_list is None, "deepmd.pt not installed"
    )
    def test_consistent(self):
        """Test consistency between different neighbor list implementations.

        Compares neighbor lists generated by dp_nblist and vesin_nblist
        to ensure they produce identical pairs and distances.
        """
        pairs_1, ds_1, _buffer_scales = dp_nblist(
            self.positions, self.box, self.nnei, self.rcut
        )
        pairs_2, ds_2, _buffer_scales = vesin_nblist(
            self.positions, self.box, self.rcut
        )
        torch.testing.assert_close(sort_pairs(pairs_1), sort_pairs(pairs_2))
        torch.testing.assert_close(torch.sort(ds_1)[0], torch.sort(ds_2)[0])



# SPDX-License-Identifier: LGPL-3.0-or-later
"""torch-admp: ADMP in PyTorch backend."""

from . import base_force, nblist, optimizer, pme, qeq, recip, spatial, utils
from ._version import __version__

__all__ = [
    "__version__",
    "base_force",
    "nblist",
    "optimizer",
    "pme",
    "qeq",
    "recip",
    "spatial",
    "utils",
]



# SPDX-License-Identifier: LGPL-3.0-or-later
import jax.numpy as jnp
import numpy as np
import openmm.app as app
import openmm.unit as unit
import torch
from dmff.api import DMFFTopology
from dmff.api.xmlio import XMLIO
from scipy import constants

from torch_admp.nblist import TorchNeighborList
from torch_admp.qeq import QEqForceModule
from torch_admp.utils import calc_grads

torch.set_default_device("cuda")
torch.set_default_dtype(torch.float64)


def load_test_data():
    xml = XMLIO()
    xml.loadXML("qeq.xml")
    res = xml.parseResidues()
    ffinfo = xml.parseXML()
    charges = [a["charge"] for a in res[0]["particles"]]
    types = np.array([a["type"] for a in res[0]["particles"]])

    # data in nm
    pdb = app.PDBFile("qeq.pdb")
    dmfftop = DMFFTopology(from_top=pdb.topology)
    positions = pdb.getPositions(asNumpy=True).value_in_unit(unit.angstrom)
    positions = jnp.array(positions)
    a, b, c = dmfftop.getPeriodicBoxVectors()

    n_atoms = dmfftop.getNumAtoms()
    eta = np.zeros([n_atoms])
    chi = np.zeros([n_atoms])
    hardness = np.zeros([n_atoms])
    for _data in ffinfo["Forces"]["ADMPQeqForce"]["node"]:
        eta[types == _data["attrib"]["type"]] = float(_data["attrib"]["eta"])
        chi[types == _data["attrib"]["type"]] = float(_data["attrib"]["chi"])
        hardness[types == _data["attrib"]["type"]] = float(_data["attrib"]["J"])

    # kJ/mol to eV/particle
    # kJ/mol to eV
    j2ev = constants.physical_constants["joule-electron volt relationship"][0]
    # kJ/mol to eV/particle
    energy_coeff = j2ev * constants.kilo / constants.Avogadro
    # length: angstrom, energy: eV
    data_dict = {
        "n_atoms": n_atoms,
        "position": np.array(positions),
        "box": np.array([a._value, b._value, c._value]) * 10.0,
        "chi": chi * energy_coeff,
        "hardness": hardness * energy_coeff,
        "eta": eta,
        "charge": charges,
    }
    return data_dict


if __name__ == "__main__":
    rcut = 8.0
    ethresh = 1e-5

    data_dict = load_test_data()

    positions = torch.tensor(
        data_dict["position"],
        requires_grad=True,
    )
    box = torch.tensor(
        data_dict["box"],
        requires_grad=False,
    )
    chi = torch.tensor(
        data_dict["chi"],
        requires_grad=False,
    )
    hardness = torch.tensor(
        data_dict["hardness"],
        requires_grad=False,
    )
    eta = torch.tensor(
        data_dict["eta"],
        requires_grad=False,
    )
    charges = torch.tensor(
        data_dict["charge"],
        requires_grad=False,
    )

    # calculate pairs
    nblist = TorchNeighborList(cutoff=rcut)
    pairs = nblist(positions, box)
    ds = nblist.get_ds()
    buffer_scales = nblist.get_buffer_scales()

    constraint_matrix = torch.ones([1, data_dict["n_atoms"]], dtype=torch.float64)
    constraint_vals = torch.zeros(1, dtype=torch.float64)
    module = QEqForceModule(rcut=rcut, ethresh=ethresh)
    energy, q_opt = module.solve_pgrad(
        charges,
        positions,
        box,
        chi,
        hardness,
        eta,
        pairs,
        ds,
        buffer_scales,
        constraint_matrix,
        constraint_vals,
    )
    forces = -calc_grads(energy, positions)

    print("QEq converges in %d step(s)" % module.converge_iter)



# SPDX-License-Identifier: LGPL-3.0-or-later
"""Tests for JIT compilation functionality in torch-admp.

This module contains tests to verify that torch-admp modules can be
successfully compiled with TorchScript JIT and produce identical results
to their non-JIT counterparts.
"""

import os
import unittest

import numpy as np
import torch

from torch_admp.env import DEVICE
from torch_admp.nblist import TorchNeighborList
from torch_admp.pme import CoulombForceModule
from torch_admp.qeq import GaussianDampingForceModule, QEqForceModule, SiteForceModule
from torch_admp.utils import calc_grads, to_numpy_array

from . import SEED

# torch.set_default_dtype(torch.float64)

rcut = 4.0
ethresh = 1e-5
l_box = 10.0
n_atoms = 100


class JITTest:
    """Test class for JIT compilation verification.

    This class provides a generic test method to compare results from
    JIT-compiled modules with their non-JIT counterparts.
    """

    def test(
        self,
    ):
        """Test JIT compilation produces identical results.

        Compares energy and gradient outputs from JIT-compiled modules
        with those from regular PyTorch modules to ensure correctness.
        """
        # Set random generators with SEED for reproducibility
        np_rng = np.random.default_rng(SEED)
        # torch_rng = torch.Generator().manual_seed(SEED)

        positions = np_rng.random((n_atoms, 3)) * l_box
        if self.periodic:
            box = np.diag([l_box, l_box, l_box])
        else:
            box = None
        charges = np_rng.uniform(-1.0, 1.0, (n_atoms))
        charges -= charges.mean()

        positions = torch.tensor(positions, requires_grad=True).to(DEVICE)
        if self.periodic:
            box = torch.tensor(box).to(DEVICE)
        charges = torch.tensor(charges, requires_grad=True).to(DEVICE)

        nblist = TorchNeighborList(cutoff=rcut)
        pairs = nblist(positions, box)
        ds = nblist.get_ds()
        buffer_scales = nblist.get_buffer_scales()

        params = {
            "charge": charges,
            "eta": torch.ones(n_atoms).to(DEVICE),
            "chi": torch.ones(n_atoms).to(DEVICE),
            "hardness": torch.zeros(n_atoms).to(DEVICE),
        }
        energy = self.module(
            positions,
            box,
            pairs,
            ds,
            buffer_scales,
            params,
        )
        jit_energy = self.jit_module(
            positions,
            box,
            pairs,
            ds,
            buffer_scales,
            params,
        )
        grad = calc_grads(energy, charges)
        jit_grad = calc_grads(jit_energy, charges)

        self.assertAlmostEqual(energy.item(), jit_energy.item())
        self.assertTrue(
            np.allclose(
                to_numpy_array(grad),
                to_numpy_array(jit_grad),
            )
        )

        torch.jit.save(self.jit_module, "./frozen_model.pth", {})

    def tearDown(self):
        """Clean up test artifacts.

        Removes any model files created during testing.
        """
        for f in os.listdir("."):
            if f == "frozen_model.pth":
                os.remove(f)


class TestOBCCoulombForceModule(unittest.TestCase, JITTest):
    """Test JIT compilation for CoulombForceModule with open boundary conditions.

    Tests that the CoulombForceModule can be JIT-compiled and produces
    identical results when using open boundary conditions.
    """

    def setUp(self):
        """Set up test for OBC CoulombForceModule JIT compilation.

        Initializes the module with open boundary conditions and creates
        a JIT-compiled version for testing.
        """
        self.periodic = False
        self.module = CoulombForceModule(rcut=rcut, ethresh=ethresh)
        self.jit_module = torch.jit.script(self.module)

    def tearDown(self):
        """Clean up test artifacts."""
        JITTest.tearDown(self)


class TestPBCCoulombForceModule(unittest.TestCase, JITTest):
    """Test JIT compilation for CoulombForceModule with periodic boundary conditions.

    Tests that the CoulombForceModule can be JIT-compiled and produces
    identical results when using periodic boundary conditions.
    """

    def setUp(self):
        """Set up test for PBC CoulombForceModule JIT compilation.

        Initializes the module with periodic boundary conditions and creates
        a JIT-compiled version for testing.
        """
        self.periodic = True
        self.module = CoulombForceModule(rcut=rcut, ethresh=ethresh)
        self.jit_module = torch.jit.script(self.module)

    def tearDown(self):
        """Clean up test artifacts."""
        JITTest.tearDown(self)


class TestSlabCorrXForceModule(unittest.TestCase, JITTest):
    """Test JIT compilation for CoulombForceModule with X-axis slab correction.

    Tests that the CoulombForceModule can be JIT-compiled and produces
    identical results when using slab correction along the X-axis.
    """

    def setUp(self):
        """Set up test for X-axis slab correction CoulombForceModule JIT compilation.

        Initializes the module with X-axis slab correction and creates
        a JIT-compiled version for testing.
        """
        self.periodic = True
        self.module = CoulombForceModule(
            rcut=rcut,
            ethresh=ethresh,
            slab_corr=True,
            slab_axis=0,
        )
        self.jit_module = torch.jit.script(self.module)

    def tearDown(self):
        """Clean up test artifacts."""
        JITTest.tearDown(self)


class TestSlabCorrYForceModule(unittest.TestCase, JITTest):
    """Test JIT compilation for CoulombForceModule with Y-axis slab correction.

    Tests that the CoulombForceModule can be JIT-compiled and produces
    identical results when using slab correction along the Y-axis.
    """

    def setUp(self):
        """Set up test for Y-axis slab correction CoulombForceModule JIT compilation.

        Initializes the module with Y-axis slab correction and creates
        a JIT-compiled version for testing.
        """
        self.periodic = True
        self.module = CoulombForceModule(
            rcut=rcut,
            ethresh=ethresh,
            slab_corr=True,
            slab_axis=1,
        )
        self.jit_module = torch.jit.script(self.module)

    def tearDown(self):
        """Clean up test artifacts."""
        JITTest.tearDown(self)


class TestSlabCorrZForceModule(unittest.TestCase, JITTest):
    """Test JIT compilation for CoulombForceModule with Z-axis slab correction.

    Tests that the CoulombForceModule can be JIT-compiled and produces
    identical results when using slab correction along the Z-axis.
    """

    def setUp(self):
        """Set up test for Z-axis slab correction CoulombForceModule JIT compilation.

        Initializes the module with Z-axis slab correction and creates
        a JIT-compiled version for testing.
        """
        self.periodic = True
        self.module = CoulombForceModule(
            rcut=rcut,
            ethresh=ethresh,
            slab_corr=True,
            slab_axis=2,
        )
        self.jit_module = torch.jit.script(self.module)

    def tearDown(self):
        """Clean up test artifacts."""
        JITTest.tearDown(self)


class TestGaussianDampingForceModule(unittest.TestCase, JITTest):
    """Test JIT compilation for GaussianDampingForceModule.

    Tests that the GaussianDampingForceModule can be JIT-compiled and produces
    identical results to its non-JIT counterpart.
    """

    def setUp(self):
        """Set up test for GaussianDampingForceModule JIT compilation.

        Initializes the module and creates a JIT-compiled version for testing.
        """
        self.periodic = True
        self.module = GaussianDampingForceModule()
        self.jit_module = torch.jit.script(self.module)

    def tearDown(self):
        """Clean up test artifacts."""
        JITTest.tearDown(self)


class TestSiteForceModule(unittest.TestCase, JITTest):
    """Test JIT compilation for SiteForceModule.

    Tests that the SiteForceModule can be JIT-compiled and produces
    identical results to its non-JIT counterpart.
    """

    def setUp(self):
        """Set up test for SiteForceModule JIT compilation.

        Initializes the module and creates a JIT-compiled version for testing.
        """
        self.periodic = True
        self.module = SiteForceModule()
        self.jit_module = torch.jit.script(self.module)

    def tearDown(self):
        """Clean up test artifacts."""
        JITTest.tearDown(self)


class TestQEqForceModule(unittest.TestCase, JITTest):
    """Test JIT compilation for QEqForceModule.

    Tests that the QEqForceModule can be JIT-compiled and produces
    identical results to its non-JIT counterpart.
    """

    def setUp(self):
        """Set up test for QEqForceModule JIT compilation.

        Initializes the module with specified cutoff and error threshold,
        and creates a JIT-compiled version for testing.
        """
        self.periodic = True
        self.module = QEqForceModule(
            rcut=rcut,
            ethresh=ethresh,
        )
        self.jit_module = torch.jit.script(self.module)

    def tearDown(self):
        """Clean up test artifacts."""
        JITTest.tearDown(self)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-3.0-or-later
import unittest
from pathlib import Path

import numpy as np
import openmm as mm
import torch
from ase import io
from openmm import app
from openmm.unit import angstrom
from scipy import constants

from torch_admp.env import DEVICE
from torch_admp.nblist import TorchNeighborList
from torch_admp.pme import CoulombForceModule
from torch_admp.utils import calc_grads, to_numpy_array

from . import SEED

# torch.set_default_dtype(torch.float64)

# kJ/mol to eV/particle
energy_coeff = (
    constants.physical_constants["joule-electron volt relationship"][0]
    * constants.kilo
    / constants.Avogadro
)
# kJ/(mol nm) to eV/particle/A
force_coeff = energy_coeff * constants.angstrom / constants.nano
# Set random generators with SEED for reproducibility
np_rng = np.random.default_rng(SEED)
torch_rng = torch.Generator(device=DEVICE).manual_seed(SEED)


class TestOpenMMSimulation:
    def __init__(self) -> None:
        self.rcut = 5.0
        self.l_box = 20.0
        self.ethresh = 5e-6
        self.n_atoms = 100

        self.charges = np_rng.uniform(-1.0, 1.0, (self.n_atoms))
        self.charges -= self.charges.mean()
        self.positions = np_rng.random((self.n_atoms, 3)) * self.l_box

    def setup(self, real_space=True):
        self.system = mm.System()
        self.system.setDefaultPeriodicBoxVectors(
            (self.l_box * angstrom, 0, 0),
            (0, self.l_box * angstrom, 0),
            (0, 0, self.l_box * angstrom),
        )
        # NonbondedForce, Particle Mesh Ewald
        nonbonded = mm.NonbondedForce()
        nonbonded.setNonbondedMethod(mm.NonbondedForce.PME)
        nonbonded.setCutoffDistance(self.rcut * angstrom)
        nonbonded.setEwaldErrorTolerance(self.ethresh)
        nonbonded.setIncludeDirectSpace(real_space)

        self.system.addForce(nonbonded)
        # add ions to the system
        for ii in range(self.n_atoms):
            self.system.addParticle(1.0)  # assume the mass is 1 for simplicity
            nonbonded.addParticle(self.charges[ii], 0, 0)

        dummy_integrator = mm.CustomIntegrator(0)
        # platform = mm.Platform.getPlatformByName("CUDA")
        # create simulation
        self.simulation = app.Simulation(
            topology=None,
            system=self.system,
            integrator=dummy_integrator,
            # platform=platform,
        )
        self.simulation.context.setPositions(self.positions * angstrom)

    def run(self):
        """
        return energy [eV], forces [eV/A]
        """
        state = self.simulation.context.getState(getEnergy=True, getForces=True)
        forces = state.getForces(asNumpy=True)
        energy = state.getPotentialEnergy()
        return (
            np.atleast_1d(energy._value)[0] * energy_coeff,
            forces._value.reshape(-1, 3) * force_coeff,
        )


class TestOBCCoulombForceModule(unittest.TestCase):
    """
    Coulomb interaction under open boundary condition
    """

    def setUp(self) -> None:
        atoms = io.read(
            str(Path(__file__).parent / "data/lmp_coul_obc/system.data"),
            format="lammps-data",
        )
        positions = atoms.get_positions()
        self.box = None
        charges = atoms.get_initial_charges()

        _positions = torch.tensor(positions, requires_grad=True)
        self.charges = torch.tensor(charges)
        self.positions = _positions.unsqueeze(0)

        self.nblist = TorchNeighborList(cutoff=4.0)
        self.pairs = self.nblist(
            self.positions.squeeze(0),
            self.box,
        ).unsqueeze(0)
        self.ds = self.nblist.get_ds().unsqueeze(0)
        self.buffer_scales = self.nblist.get_buffer_scales().unsqueeze(0)

        self.module = CoulombForceModule(rcut=5.0, ethresh=1e-5)
        # test jit-able
        self.jit_module = torch.jit.script(self.module)

    def test_numerical(self):
        ref_energy = np.loadtxt(
            str(Path(__file__).parent / "data/lmp_coul_obc/thermo.out")
        ).reshape(-1)[1]
        ref_atoms = io.read(
            str(Path(__file__).parent / "data/lmp_coul_obc/dump.lammpstrj")
        )
        ref_forces = ref_atoms.get_forces()

        energy = self.module(
            self.positions,
            self.box,
            self.pairs,
            self.ds,
            self.buffer_scales,
            {"charge": self.charges},
        )
        jit_energy = self.jit_module(
            self.positions,
            self.box,
            self.pairs,
            self.ds,
            self.buffer_scales,
            {"charge": self.charges},
        )
        forces = -calc_grads(energy, self.positions)
        jit_forces = -calc_grads(jit_energy, self.positions)

        # energy [eV]
        for e in [energy, jit_energy]:
            np.testing.assert_allclose(
                to_numpy_array(e),
                [ref_energy],
                atol=1e-6,
                rtol=1e-6,
            )
        # force [eV/A]
        for f in [forces, jit_forces]:
            np.testing.assert_allclose(
                to_numpy_array(f).reshape(-1, 3),
                ref_forces,
                atol=1e-6,
                rtol=1e-6,
            )


class TestPBCCoulombForceModule(unittest.TestCase):
    def setUp(self) -> None:
        self.ref_system = TestOpenMMSimulation()
        self.ref_system.setup(real_space=True)

        _positions = torch.tensor(self.ref_system.positions, requires_grad=True)
        self.charges = torch.tensor(self.ref_system.charges).unsqueeze(0)
        _box = torch.tensor(
            np.diag(
                [self.ref_system.l_box, self.ref_system.l_box, self.ref_system.l_box]
            )
        )
        self.positions = _positions.unsqueeze(0)
        self.box = _box.unsqueeze(0)

        self.nblist = TorchNeighborList(cutoff=self.ref_system.rcut)
        self.pairs = self.nblist(
            self.positions.squeeze(0),
            self.box.squeeze(0),
        ).unsqueeze(0)
        self.ds = self.nblist.get_ds().unsqueeze(0)
        self.buffer_scales = self.nblist.get_buffer_scales().unsqueeze(0)

        self.module = CoulombForceModule(
            rcut=self.ref_system.rcut,
            ethresh=self.ref_system.ethresh,
        )
        # test jit-able
        self.jit_module = torch.jit.script(self.module)

    def test_numerical(self):
        energy = self.module(
            self.positions,
            self.box,
            self.pairs,
            self.ds,
            self.buffer_scales,
            {"charge": self.charges},
        )
        jit_energy = self.jit_module(
            self.positions,
            self.box,
            self.pairs,
            self.ds,
            self.buffer_scales,
            {"charge": self.charges},
        )
        forces = -calc_grads(energy, self.positions)
        jit_forces = -calc_grads(jit_energy, self.positions)

        nonbonded = self.ref_system.system.getForce(0)
        # A^-1 to nm^-1 for kappa
        nonbonded.setPMEParameters(
            self.module.kappa * 10.0,
            self.module.kmesh[0, 0].item(),
            self.module.kmesh[0, 1].item(),
            self.module.kmesh[0, 2].item(),
        )
        # simulation = self.ref_system.simulation
        # ewald_params = nonbonded.getPMEParametersInContext(simulation.context)
        # self.kappa = ewald_params[0] / 10.0
        # self.kmesh = tuple(ewald_params[1:])
        ref_energy, ref_forces = self.ref_system.run()

        # energy [eV]
        for e in [energy, jit_energy]:
            np.testing.assert_allclose(
                to_numpy_array(e),
                [ref_energy],
                atol=1e-5,
                rtol=1e-5,
            )
        # force [eV/A]
        for f in [forces, jit_forces]:
            np.testing.assert_allclose(
                to_numpy_array(f).reshape(-1, 3),
                ref_forces,
                atol=1e-5,
                rtol=1e-5,
            )


class TestPBCSlabCorrCoulombForceModule(unittest.TestCase):
    def setUp(self) -> None:
        atoms = io.read(
            str(Path(__file__).parent / "data/lmp_coul_pbc/system.data"),
            format="lammps-data",
        )
        positions = atoms.get_positions()
        box = atoms.get_cell().array
        charges = atoms.get_initial_charges()

        _positions = torch.tensor(positions, requires_grad=True)
        _box = torch.tensor(box)
        self.charges = torch.tensor(charges).unsqueeze(0)
        self.positions = _positions.unsqueeze(0)
        self.box = _box.unsqueeze(0)

        self.nblist = TorchNeighborList(cutoff=4.0)
        self.pairs = self.nblist(
            self.positions.squeeze(0), self.box.squeeze(0)
        ).unsqueeze(0)
        self.ds = self.nblist.get_ds().unsqueeze(0)
        self.buffer_scales = self.nblist.get_buffer_scales().unsqueeze(0)

    def lammps_ref_data(self):
        e1 = np.loadtxt(
            str(Path(__file__).parent / "data/lmp_coul_pbc/thermo.out")
        ).reshape(-1)[1]
        e2 = np.loadtxt(
            str(Path(__file__).parent / "data/lmp_coul_pbc_slab_corr/thermo.out")
        ).reshape(-1)[1]
        ref_energy = e2 - e1

        atoms = io.read(str(Path(__file__).parent / "data/lmp_coul_pbc/dump.lammpstrj"))
        f1 = atoms.get_forces()
        atoms = io.read(
            str(Path(__file__).parent / "data/lmp_coul_pbc_slab_corr/dump.lammpstrj")
        )
        f2 = atoms.get_forces()
        ref_force = f2 - f1
        return ref_energy, ref_force

    def test_numerical(self):
        # rcut and ethresh are not used in slab correction calculation
        module = CoulombForceModule(
            rcut=self.nblist.cutoff,
            ethresh=1e-3,
            kspace=False,
            rspace=False,
            slab_corr=True,
            slab_axis=2,
        )
        jit_module = torch.jit.script(module)

        energy = module(
            self.positions,
            self.box,
            self.pairs,
            self.ds,
            self.buffer_scales,
            {"charge": self.charges},
        )
        jit_energy = jit_module(
            self.positions,
            self.box,
            self.pairs,
            self.ds,
            self.buffer_scales,
            {"charge": self.charges},
        )
        forces = -calc_grads(energy, self.positions)
        jit_forces = -calc_grads(jit_energy, self.positions)

        # ref_energy = self.make_ref_data(axis=2)
        ref_energy, ref_forces = self.lammps_ref_data()

        # energy [eV]
        for e in [energy, jit_energy]:
            np.testing.assert_allclose(
                to_numpy_array(e),
                [ref_energy],
                atol=1e-6,
                rtol=1e-6,
            )
        # force [eV/A]
        for f in [forces, jit_forces]:
            np.testing.assert_allclose(
                to_numpy_array(f).reshape(-1, 3),
                ref_forces,
                atol=1e-6,
                rtol=1e-6,
            )


class TestCoulombForceModule(unittest.TestCase):
    """Test cases to improve coverage of CoulombForceModule"""

    def setUp(self) -> None:
        # Setup for getter tests
        self.module = CoulombForceModule(rcut=5.0, ethresh=1e-5, sel=[10, 20])

        # Setup for edge case tests
        self.n_atoms = 10
        self.positions = torch.rand(1, self.n_atoms, 3, generator=torch_rng) * 10.0
        self.box = torch.diag(torch.tensor([10.0, 10.0, 10.0])).unsqueeze(0)
        self.charges = torch.randn(1, self.n_atoms, generator=torch_rng)

        self.nblist = TorchNeighborList(cutoff=4.0)
        self.pairs = self.nblist(
            self.positions.squeeze(0), self.box.squeeze(0)
        ).unsqueeze(0)
        self.ds = self.nblist.get_ds().unsqueeze(0)
        self.buffer_scales = self.nblist.get_buffer_scales().unsqueeze(0)

    def test_get_rcut(self):
        """Test get_rcut method (line 98)"""
        self.assertEqual(self.module.get_rcut(), 5.0)

    def test_get_sel(self):
        """Test get_sel method (line 101)"""
        self.assertEqual(self.module.get_sel(), [10, 20])

    def test_kspace_false(self):
        """Test _forward_pbc_self when kspace_flag is False (line 288)"""
        module = CoulombForceModule(rcut=5.0, ethresh=1e-5, kspace=False)
        module(
            self.positions,
            self.box,
            self.pairs,
            self.ds,
            self.buffer_scales,
            {"charge": self.charges},
        )
        # Verify that self_energy is zero when kspace_flag is False
        self.assertEqual(module.self_energy.item(), 0.0)
        self.assertEqual(module.reciprocal_energy.item(), 0.0)

    def test_setup_ewald_parameters_openmm(self):
        """Test setup_ewald_parameters with openmm method (lines 388-443)"""
        from torch_admp.pme import setup_ewald_parameters

        box = np.diag([10.0, 10.0, 10.0])
        kappa, kx, ky, kz = setup_ewald_parameters(
            rcut=5.0, box=box, threshold=1e-5, method="openmm"
        )

        # Verify that parameters are reasonable
        self.assertGreater(kappa, 0)
        self.assertGreaterEqual(kx, 1)
        self.assertGreaterEqual(ky, 1)
        self.assertGreaterEqual(kz, 1)

    def test_setup_ewald_parameters_gromacs(self):
        """Test setup_ewald_parameters with gromacs method (lines 388-443)"""
        from torch_admp.pme import setup_ewald_parameters

        box = np.diag([10.0, 10.0, 10.0])
        kappa, kx, ky, kz = setup_ewald_parameters(
            rcut=5.0, box=box, threshold=1e-5, spacing=1.0, method="gromacs"
        )

        # Verify that parameters are reasonable
        self.assertGreater(kappa, 0)
        self.assertGreaterEqual(kx, 1)
        self.assertGreaterEqual(ky, 1)
        self.assertGreaterEqual(kz, 1)

    def test_setup_ewald_parameters_no_box(self):
        """Test setup_ewald_parameters with no box (lines 388-443)"""
        from torch_admp.pme import setup_ewald_parameters

        kappa, kx, ky, kz = setup_ewald_parameters(rcut=5.0, box=None)

        # Should return default values
        self.assertEqual(kappa, 0.1)
        self.assertEqual(kx, 1)
        self.assertEqual(ky, 1)
        self.assertEqual(kz, 1)

    def test_setup_ewald_parameters_invalid_method(self):
        """Test setup_ewald_parameters with invalid method (lines 388-443)"""
        from torch_admp.pme import setup_ewald_parameters

        box = np.diag([10.0, 10.0, 10.0])

        with self.assertRaises(ValueError):
            setup_ewald_parameters(rcut=5.0, box=box, threshold=1e-5, method="invalid")

    def test_setup_ewald_parameters_gromacs_no_spacing(self):
        """Test setup_ewald_parameters with gromacs method but no spacing (lines 388-443)"""
        from torch_admp.pme import setup_ewald_parameters

        box = np.diag([10.0, 10.0, 10.0])

        with self.assertRaises(AssertionError):
            setup_ewald_parameters(rcut=5.0, box=box, threshold=1e-5, method="gromacs")

    def test_setup_ewald_parameters_non_orthogonal_box(self):
        """Test setup_ewald_parameters with non-orthogonal box (lines 388-443)"""
        from torch_admp.pme import setup_ewald_parameters

        # Create a non-orthogonal box
        box = np.array([[10.0, 1.0, 0.0], [0.0, 10.0, 0.0], [0.0, 0.0, 10.0]])

        with self.assertRaises(AssertionError):
            setup_ewald_parameters(rcut=5.0, box=box, threshold=1e-5, method="openmm")



# SPDX-License-Identifier: LGPL-3.0-or-later
"""
Neighbor list utilities for torch-admp.

This module provides functions and classes for building and managing neighbor lists
used in molecular simulations, including implementations for both periodic and
non-periodic boundary conditions.
"""

import warnings
from typing import Optional, Tuple

import torch

try:
    from deepmd.pt.utils.nlist import extend_input_and_build_neighbor_list
except ImportError:
    warnings.warn("deepmd.pt is required for dp_nblist", ImportWarning)
    extend_input_and_build_neighbor_list = None

from vesin.torch import NeighborList

from torch_admp.env import DEVICE, GLOBAL_PT_FLOAT_PRECISION
from torch_admp.spatial import pbc_shift


def dp_nblist(
    positions: torch.Tensor,
    box: Optional[torch.Tensor],
    nnei: int,
    rcut: float,
):
    """
    Build neighbor list data based on DP (Deep Potential) functions.

    Parameters
    ----------
    positions : torch.Tensor
        Atomic positions
    box : Optional[torch.Tensor]
        Simulation box vectors
    nnei : int
        Number of neighbors
    rcut : float
        Cutoff radius

    Returns
    -------
    tuple
        Tuple containing (pairs, ds, buffer_scales)

    Raises
    ------
    ImportError
        If deepmd.pt is not available
    """
    if extend_input_and_build_neighbor_list is None:
        raise ImportError(
            "deepmd.pt is required for dp_nblist. Please install deepmd-pt to use this function."
        )

    positions = torch.reshape(positions, [1, -1, 3])
    (
        extended_coord,
        extended_atype,
        mapping,
        nlist,
    ) = extend_input_and_build_neighbor_list(
        positions,
        torch.zeros(1, positions.shape[1]),
        rcut,
        [nnei],
        box=box,
    )
    extended_pairs = make_extended_pairs(nlist)
    pairs, _buffer_scales, mask_ij, mask_ii = make_local_pairs(extended_pairs, mapping)
    buffer_scales = _buffer_scales.to(positions.device)
    ds_ij = make_ds(extended_pairs, extended_coord, mask_ij)
    ds_ii = make_ds(extended_pairs, extended_coord, mask_ii)
    ds = torch.concat([ds_ij, ds_ii])
    del extended_coord, extended_atype
    return pairs, ds, buffer_scales


def vesin_nblist(
    positions: torch.Tensor,
    box: Optional[torch.Tensor],
    rcut: float,
):
    """
    Build neighbor list using the Vesin library.

    Parameters
    ----------
    positions : torch.Tensor
        Atomic positions
    box : Optional[torch.Tensor]
        Simulation box vectors
    rcut : float
        Cutoff radius

    Returns
    -------
    tuple
        Tuple containing (pairs, ds, buffer_scales)
    """
    device = positions.device
    calculator = NeighborList(cutoff=rcut, full_list=False)

    # Handle the box parameter properly
    box_cpu = box.to("cpu") if box is not None else None

    # Use type ignore to work around the type checking issue
    ii, jj, ds = calculator.compute(
        points=positions.to("cpu"),
        box=box_cpu,  # type: ignore
        periodic=True,
        quantities="ijd",
    )
    buffer_scales = torch.ones_like(ds).to(device)
    return torch.stack([ii, jj]).to(device).T, ds.to(device), buffer_scales


def make_extended_pairs(
    nlist: torch.Tensor,
) -> torch.Tensor:
    """Return the pairs between local and extended indices.

    Parameters
    ----------
    nlist : torch.Tensor
        nframes x nloc x nsel, neighbor list between local and extended indices

    Returns
    -------
    extended_pairs: torch.Tensor
        [[i1, j1], [i2, j2], ...],
        in which i is the local index and j is the extended index
    """
    nframes, nloc, nsel = nlist.shape
    assert nframes == 1
    nlist_reshape = torch.reshape(nlist, [nframes, nloc * nsel, 1])
    # nlist is padded with -1
    mask = nlist_reshape.ge(0)

    ii = torch.arange(nloc, dtype=torch.int64, device=nlist.device)
    ii = torch.tile(ii.reshape(-1, 1), [1, nsel])
    ii = torch.reshape(ii, [nframes, nloc * nsel, 1])
    sel_ii = torch.masked_select(ii, mask)

    # nf x (nloc x nsel)
    sel_jj = torch.masked_select(nlist_reshape, mask)
    extended_pairs = torch.stack([sel_ii, sel_jj], dim=-1)
    return extended_pairs


def make_local_pairs(
    extended_pairs: torch.Tensor,
    mapping: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """Return the pairs between local indices.

    Parameters
    ----------
    extended_pairs : torch.Tensor
        npairs_all x 2,
    mapping : torch.Tensor
        nframes x nall, index from extended to local

    Returns
    -------
    local_pairs: torch.Tensor
        npairs_loc x 2, [[i1, j1], [i2, j2], ...],
        in which i and j are the local indices of the atoms (i < j)
    mask: torch.Tensor
        npairs_all, mask for the local pairs (i < j)
    """
    nframes, _nall = mapping.shape
    assert nframes == 1
    ii = extended_pairs[..., 0]
    jj = torch.gather(mapping.reshape(-1), 0, extended_pairs[..., 1])

    mask_ij = ii.lt(jj)
    mask_ii = ii.eq(jj)
    local_pairs_ij = torch.stack([ii, jj], dim=-1)[mask_ij]
    local_pairs_ii = torch.stack([ii, jj], dim=-1)[mask_ii]

    buffer_scales_ij = torch.ones(local_pairs_ij.shape[0], device=local_pairs_ij.device)
    buffer_scales_ii = (
        torch.ones(local_pairs_ii.shape[0], device=local_pairs_ii.device) / 2.0
    )

    local_pairs = torch.concat([local_pairs_ij, local_pairs_ii])
    buffer_scales = torch.concat([buffer_scales_ij, buffer_scales_ii])
    return local_pairs, buffer_scales, mask_ij, mask_ii


def make_ds(
    extended_pairs: torch.Tensor,
    extended_coord: torch.Tensor,
    pairs_mask: torch.Tensor,
) -> torch.Tensor:
    """Calculate the i-j distance from the neighbor list.

    Parameters
    ----------
    extended_pairs : torch.Tensor
        npairs_all x 2,
    extended_coord : torch.Tensor
        nframes x nall x 3, extended coordinates
    pairs_mask : torch.Tensor
        npairs_all, mask for the local pairs (i < j)

    Returns
    -------
    ds: torch.Tensor
        npairs_loc, i-j distance
    """
    nframes, _nall, _ = extended_coord.shape
    assert nframes == 1

    ii = extended_pairs[..., 0]
    jj = extended_pairs[..., 1]
    diff = extended_coord[:, jj] - extended_coord[:, ii]
    ds = torch.norm(diff.reshape(-1, 3)[pairs_mask], dim=-1)
    return ds


def sort_pairs(pairs: torch.Tensor) -> torch.Tensor:
    """
    Sort atom pairs lexicographically.

    Sorts pairs first by the first index, then by the second index.

    Parameters
    ----------
    pairs : torch.Tensor
        Tensor of atom pairs

    Returns
    -------
    torch.Tensor
        Sorted tensor of atom pairs
    """
    indices = torch.argsort(pairs[:, 1])
    pairs = pairs[indices]
    indices = torch.argsort(pairs[:, 0], stable=True)
    sorted_pairs = pairs[indices]
    return sorted_pairs


class TorchNeighborList(torch.nn.Module):
    """
    Torch-compatible neighbor list implementation.

    Adapted from the curator library for JIT compatibility:
    https://github.com/Yangxinsix/curator/tree/master
    curator.data.TorchNeighborList
    """

    def __init__(
        self,
        cutoff: float,
    ) -> None:
        """
        Initialize the TorchNeighborList.

        Parameters
        ----------
        cutoff : float
            Cutoff distance for neighbor list construction
        """
        super().__init__()
        self.cutoff = cutoff
        _t = torch.arange(-1, 2, device=DEVICE)
        disp_mat = torch.cartesian_prod(_t, _t, _t)
        self.register_buffer("disp_mat", disp_mat, persistent=True)

        self.pairs = torch.jit.annotate(torch.Tensor, torch.empty(1, dtype=torch.long))
        self.buffer_scales = torch.jit.annotate(
            torch.Tensor, torch.empty(1, dtype=torch.long)
        )
        self.ds = torch.jit.annotate(
            torch.Tensor, torch.empty(1, dtype=GLOBAL_PT_FLOAT_PRECISION)
        )

    def forward(
        self, positions: torch.Tensor, box: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Compute neighbor list for given positions.

        Parameters
        ----------
        positions : torch.Tensor
            Atomic positions
        box : Optional[torch.Tensor], optional
            Simulation box vectors, by default None

        Returns
        -------
        torch.Tensor
            Tensor of atom pairs
        """
        if box is None:
            pairs = self.forward_obc(positions)
            pbc_flag = False
        else:
            check_cutoff(box, self.cutoff)
            pairs = self.forward_pbc(positions, box)
            pbc_flag = True

        self.pairs = pairs
        self.buffer_scales = self.pairs_buffer_scales(pairs)
        self.ds = self.pairs_ds(positions, pairs, box, pbc_flag)
        return pairs

    def forward_pbc(
        self,
        positions: torch.Tensor,
        box: torch.Tensor,
    ) -> torch.Tensor:
        """
        Compute neighbor list for periodic boundary conditions.

        Parameters
        ----------
        positions : torch.Tensor
            Atomic positions
        box : torch.Tensor
            Simulation box vectors

        Returns
        -------
        torch.Tensor
            Tensor of atom pairs
        """
        # calculate padding size. It is useful for all kinds of cells
        wrapped_pos = self.wrap_positions(positions, box)
        norm_a = torch.cross(box[1], box[2]).norm()
        norm_b = torch.cross(box[2], box[0]).norm()
        norm_c = torch.cross(box[0], box[1]).norm()
        volume = torch.sum(box[0] * torch.cross(box[1], box[2]))

        # get padding size and padding matrix to generate padded atoms. Use minimal image convention
        padding_a = torch.ceil(self.cutoff * norm_a / volume).long()
        padding_b = torch.ceil(self.cutoff * norm_b / volume).long()
        padding_c = torch.ceil(self.cutoff * norm_c / volume).long()

        padding_mat = torch.cartesian_prod(
            torch.arange(
                -padding_a.item(), padding_a.item() + 1, device=padding_a.device
            ),
            torch.arange(
                -padding_b.item(), padding_b.item() + 1, device=padding_a.device
            ),
            torch.arange(
                -padding_c.item(), padding_c.item() + 1, device=padding_a.device
            ),
        ).to(box.dtype)
        padding_size = (2 * padding_a + 1) * (2 * padding_b + 1) * (2 * padding_c + 1)

        # padding, calculating box numbers and shapes
        padded_pos = (wrapped_pos.unsqueeze(1) + padding_mat @ box).view(-1, 3)
        padded_cpos = torch.floor(padded_pos / self.cutoff).long()
        corner = torch.min(padded_cpos, dim=0)[0]  # the box at the corner
        padded_cpos -= corner
        c_pos_shap = torch.max(padded_cpos, dim=0)[0] + 1  # c_pos starts from 0
        num_cells = int(torch.prod(c_pos_shap).item())
        count_vec = torch.ones_like(c_pos_shap)
        count_vec[0] = c_pos_shap[1] * c_pos_shap[2]
        count_vec[1] = c_pos_shap[2]

        padded_cind = torch.sum(padded_cpos * count_vec, dim=1)
        padded_gind = (
            torch.arange(padded_cind.shape[0], device=count_vec.device) + 1
        )  # global index of padded atoms, starts from 1
        padded_rind = torch.arange(
            positions.shape[0], device=count_vec.device
        ).repeat_interleave(padding_size)  # local index of padded atoms in the unit box

        # atom box position and index
        atom_cpos = torch.floor(wrapped_pos / self.cutoff).long() - corner
        # atom neighbors' box position and index
        # Ensure disp_mat is on the same device as atom_cpos
        # Use type: ignore to work around type checking issue with registered buffers
        disp_mat_device = self.disp_mat.to(atom_cpos.device)  # type: ignore
        atom_cnpos = atom_cpos.unsqueeze(1) + disp_mat_device  # type: ignore
        atom_cnind = torch.sum(atom_cnpos * count_vec, dim=-1)

        # construct a C x N matrix to store the box atom list, this is the most expensive part.
        padded_cind_sorted, padded_cind_args = torch.sort(padded_cind, stable=True)
        cell_ind, cell_atom_num = torch.unique_consecutive(
            padded_cind_sorted, return_counts=True
        )
        max_cell_anum = int(cell_atom_num.max().item())
        global_cell_ind = torch.zeros(
            (num_cells, max_cell_anum, 2),
            dtype=c_pos_shap.dtype,
            device=c_pos_shap.device,
        )
        cell_aind = torch.nonzero(
            torch.arange(max_cell_anum, device=count_vec.device).repeat(
                cell_atom_num.shape[0], 1
            )
            < cell_atom_num.unsqueeze(-1)
        )[:, 1]
        global_cell_ind[padded_cind_sorted, cell_aind, 0] = padded_gind[
            padded_cind_args
        ]
        global_cell_ind[padded_cind_sorted, cell_aind, 1] = padded_rind[
            padded_cind_args
        ]

        # masking
        atom_nind = global_cell_ind[atom_cnind]
        pair_i, neigh, j = torch.where(atom_nind[:, :, :, 0])
        pair_j = atom_nind[pair_i, neigh, j, 1]
        pair_j_padded = (
            atom_nind[pair_i, neigh, j, 0] - 1
        )  # remember global index of padded atoms starts from 1
        pair_diff = padded_pos[pair_j_padded] - wrapped_pos[pair_i]
        pair_dist = torch.norm(pair_diff, dim=1)
        mask = torch.logical_and(
            pair_dist < self.cutoff, pair_dist > 0.01
        )  # 0.01 for numerical stability
        pairs = torch.hstack((pair_i.unsqueeze(-1), pair_j.unsqueeze(-1)))
        return pairs[mask].to(torch.long)

    def wrap_positions(
        self,
        positions: torch.Tensor,
        box: torch.Tensor,
    ) -> torch.Tensor:
        """
        Wrap positions into the unit cell.

        Parameters
        ----------
        positions : torch.Tensor
            Atomic positions
        box : torch.Tensor
            Simulation box vectors

        Returns
        -------
        torch.Tensor
            Wrapped positions
        """
        eps = torch.tensor(1e-7, device=positions.device, dtype=positions.dtype)
        # wrap atoms outside of the box
        scaled_pos = (positions @ torch.linalg.inv(box) + eps) % 1.0 - eps
        return scaled_pos @ box

    def forward_obc(self, positions: torch.Tensor) -> torch.Tensor:
        """
        Compute neighbor list for open boundary conditions.

        Parameters
        ----------
        positions : torch.Tensor
            Atomic positions

        Returns
        -------
        torch.Tensor
            Tensor of atom pairs
        """
        dist_mat = torch.cdist(positions, positions)
        mask = dist_mat < self.cutoff
        mask.fill_diagonal_(False)
        pairs = torch.argwhere(mask)
        return pairs.to(torch.long)

    @staticmethod
    def pairs_buffer_scales(pairs: torch.Tensor) -> torch.Tensor:
        """
        Calculate buffer scales for atom pairs.

        Returns 1 if pair_i < pair_j, else 0.
        Used to exclude repeated pairs and buffer pairs.

        Parameters
        ----------
        pairs : torch.Tensor
            Tensor of atom pairs

        Returns
        -------
        torch.Tensor
            Buffer scales for each pair
        """
        dp = pairs[:, 0] - pairs[:, 1]
        return torch.where(
            dp < 0,
            torch.tensor(1, dtype=torch.long, device=pairs.device),
            torch.tensor(0, dtype=torch.long, device=pairs.device),
        )

    @staticmethod
    def pairs_ds(
        positions: torch.Tensor,
        pairs: torch.Tensor,
        box: Optional[torch.Tensor] = None,
        pbc_flag: bool = True,
    ) -> torch.Tensor:
        """
        Calculate distances between atom pairs.

        Parameters
        ----------
        positions : torch.Tensor
            Atomic positions
        pairs : torch.Tensor
            Tensor of atom pairs
        box : Optional[torch.Tensor], optional
            Simulation box vectors, by default None
        pbc_flag : bool, optional
            Whether to apply periodic boundary conditions, by default True

        Returns
        -------
        torch.Tensor
            Distances between atom pairs
        """
        ri = positions[pairs[:, 0]]
        rj = positions[pairs[:, 1]]
        if pbc_flag is False:
            dr = rj - ri
        else:
            assert box is not None, "Box should be provided for periodic system."
            dr = pbc_shift(ri - rj, box)
        ds = torch.norm(dr, dim=1)
        return ds

    def set_pairs(self, pairs: torch.Tensor) -> None:
        """
        Set the atom pairs.

        Parameters
        ----------
        pairs : torch.Tensor
            Tensor of atom pairs
        """
        self.pairs = pairs

    def set_buffer_scales(self, buffer_scales: torch.Tensor) -> None:
        """
        Set the buffer scales for atom pairs.

        Parameters
        ----------
        buffer_scales : torch.Tensor
            Buffer scales for each pair
        """
        self.buffer_scales = buffer_scales

    def set_ds(self, ds: torch.Tensor) -> None:
        """
        Set the distances between atom pairs.

        Parameters
        ----------
        ds : torch.Tensor
            Distances between atom pairs
        """
        self.ds = ds

    def get_pairs(self) -> torch.Tensor:
        """
        Get the atom pairs.

        Returns
        -------
        torch.Tensor
            Tensor of atom pairs
        """
        return self.pairs

    def get_buffer_scales(self) -> torch.Tensor:
        """
        Get the buffer scales for atom pairs.

        Returns
        -------
        torch.Tensor
            Buffer scales for each pair
        """
        return self.buffer_scales

    def get_ds(self) -> torch.Tensor:
        """
        Get the distances between atom pairs.

        Returns
        -------
        torch.Tensor
            Distances between atom pairs
        """
        return self.ds


def check_cutoff(box: torch.Tensor, cutoff: float) -> None:
    """
    Check whether the sphere of cutoff radius is inside the box.

    Parameters
    ----------
    box : torch.Tensor
        Simulation box vectors
    cutoff : float
        Cutoff radius

    Raises
    ------
    AssertionError
        If cutoff is larger than half the minimum height of the box
    """
    # Get the three cell vectors a1, a2, a3
    a1, a2, a3 = box[0], box[1], box[2]

    # Compute normals to the three faces
    normals = torch.stack(
        [torch.cross(a2, a3), torch.cross(a3, a1), torch.cross(a1, a2)]
    )  # shape (3, 3)

    # Normalize normals
    unit_normals = normals / torch.norm(normals, dim=1, keepdim=True)

    # Heights from origin to the faces (dot of ai with corresponding normal)
    heights = torch.abs(torch.einsum("ij,ij->i", box, unit_normals))  # shape (3,)

    # Minimum half-height (distance from origin to nearest face along normal direction)
    min_half_height = torch.min(heights) / 2

    assert cutoff <= min_half_height, (
        f"Cutoff {cutoff} is larger than half the minimum height {min_half_height} of the box. "
        "This may lead to unphysical results."
    )
